# Copyright 2023 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Binary operator override class for Tensor overrides."""
import numbers
import numpy as np

from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor as tensor_lib
from tensorflow.python.framework import tensor_shape
from tensorflow.python.ops import gen_math_ops
from tensorflow.python.ops.numpy_ops import np_dtypes
from tensorflow.python.util import nest
from tensorflow.python.util import traceback_utils


def _maybe_get_dtype(x):
  """Returns a numpy type if available from x. Skips if x is numpy.ndarray."""
  # Don't put np.ndarray in this list, because np.result_type looks at the
  # value (not just dtype) of np.ndarray to decide the result type.
  if isinstance(x, numbers.Real):
    return x
  if isinstance(x, tensor_lib.Tensor):
    return x.dtype.as_numpy_dtype
  if isinstance(x, dtypes.DType):
    return x.as_numpy_dtype
  if isinstance(x, tensor_shape.TensorShape):
    return np.int32
  if isinstance(x, (list, tuple)):
    raise ValueError(f"Cannot determine dtype.  Got sequence {x}.")
  return x


def maybe_promote_tensors(*tensors, force_same_dtype=False):
  """Promotes tensors if numpy style promotion is enabled.

  This function promotes `tensors` according to numpy promotion rules
  if numpy style promotion is enabled.  Otherwise, if
  `force_same_dtype` is `True`, it force-casts `tensors[1:]` to
  `tensor[0]`'s dtype. Note that this force-cast can be problematic.
  For example, when some `tensors[1:]` elements can be silently
  downcasted.

  Args:
    *tensors: the list of tensors to promote.
    force_same_dtype: bool (optional, default to `False`). When numpy
      style promotion is disabled and `force_same_dtype` is `True`,
      this function will force-casts `tensors[1:]` to `tensor[0]`'s
      dtype (which could be problematic).

  Returns:
    The promoted list of tensors.
  """
  if ops.is_auto_dtype_conversion_enabled():
    return tensors
  if not tensors:
    return tensors
  if not ops.is_numpy_style_type_promotion():
    if not force_same_dtype:
      return tensors
    promoted_tensors = []
    promoted_tensors.append(tensors[0])
    dtype = tensors[0].dtype.base_dtype
    for tensor in tensors[1:]:
      promoted_tensors.append(
          ops.convert_to_tensor(tensor, dtype, name="x"))
    return promoted_tensors
  result_type = np_dtypes._result_type(  # pylint: disable=protected-access
      *[_maybe_get_dtype(x) for x in nest.flatten(tensors)])
  def _promote_or_cast(x):
    if isinstance(x, tensor_lib.Tensor):
      x = gen_math_ops.cast(x, result_type)
    else:
      x = ops.convert_to_tensor(x, result_type)
    return x
  return [_promote_or_cast(x) for x in tensors]


# pylint: disable=protected-access
def override_binary_operator_helper(
    func, op_name, clazz_object=tensor_lib.Tensor):
  """Register operators with different tensor and scalar versions.

  If `clazz_object` is `SparseTensor`, assumes `func` takes `(sp_indices,
  sp_values, sp_shape, dense)` and outputs `(new_sp_values)`.

  Args:
    func: the operator
    op_name: name of the operator being overridden
    clazz_object: class to override for.  Either `Tensor` or `SparseTensor`.
  """

  @traceback_utils.filter_traceback
  def binary_op_wrapper(x, y):
    with ops.name_scope(None, op_name, [x, y]) as name:
      try:
        # force_same_dtype=False to preserve existing TF behavior
        # TODO(b/178860388): Figure out why binary_op_wrapper and
        #   r_binary_op_wrapper use different force_same_dtype values.
        x, y = maybe_promote_tensors(x, y)
        return func(x, y, name=name)
      except (TypeError, ValueError) as e:
        # Even if dispatching the op failed, the RHS may be a tensor aware
        # object that can implement the operator with knowledge of itself
        # and the tensor.
        # If the RHS is not tensor aware we still want to raise the
        # original error from the LHS, because it may be more
        # informative.
        if hasattr(type(y), "__r%s__" % op_name):
          try:
            r_op = getattr(y, "__r%s__" % op_name)
            out = r_op(x)
            if out is NotImplemented:
              raise
            return out
          except (TypeError, ValueError):
            raise e
        else:
          raise

  @traceback_utils.filter_traceback
  def binary_op_wrapper_sparse(sp_x, y):
    with ops.name_scope(None, op_name, [sp_x, y]) as name:
      y = ops.convert_to_tensor(y, dtype=sp_x.dtype.base_dtype, name="y")
      # use the passed-in SparseTensor class to avoid having to import
      # SparseTensor, which would cause a cyclic dep with math_ops
      return clazz_object(
          sp_x.indices,
          func(sp_x.indices, sp_x.values, sp_x.dense_shape, y, name=name),
          sp_x.dense_shape)

  @traceback_utils.filter_traceback
  def r_binary_op_wrapper(y, x):
    with ops.name_scope(None, op_name, [x, y]) as name:
      # TODO(b/178860388): Figure out why binary_op_wrapper and
      #   r_binary_op_wrapper use different force_same_dtype values.
      y, x = maybe_promote_tensors(y, x, force_same_dtype=True)
      return func(x, y, name=name)

  # Propagate func.__doc__ to the wrappers
  try:
    doc = func.__doc__
  except AttributeError:
    doc = None
  binary_op_wrapper.__doc__ = doc
  r_binary_op_wrapper.__doc__ = doc
  binary_op_wrapper_sparse.__doc__ = doc

  if clazz_object is tensor_lib.Tensor:
    clazz_object._override_operator("__%s__" % op_name, binary_op_wrapper)
    del binary_op_wrapper
    clazz_object._override_operator("__r%s__" % op_name, r_binary_op_wrapper)
    del r_binary_op_wrapper
  else:
    clazz_object._override_operator("__%s__" % op_name,
                                    binary_op_wrapper_sparse)
    del binary_op_wrapper_sparse



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tensorflow/compiler/mlir/lite/metrics/converter_error_data.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'tensorflow/compiler/mlir/lite/metrics/converter_error_data.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n@tensorflow/compiler/mlir/lite/metrics/converter_error_data.proto\x12\x0etflite.metrics\"\xdc\x06\n\x12\x43onverterErrorData\x12\x11\n\tcomponent\x18\x01 \x01(\t\x12\x14\n\x0csubcomponent\x18\x02 \x01(\t\x12@\n\nerror_code\x18\x03 \x01(\x0e\x32,.tflite.metrics.ConverterErrorData.ErrorCode\x12\x15\n\rerror_message\x18\x04 \x01(\t\x12=\n\x08operator\x18\x05 \x01(\x0b\x32+.tflite.metrics.ConverterErrorData.Operator\x12=\n\x08location\x18\x06 \x01(\x0b\x32+.tflite.metrics.ConverterErrorData.Location\x1a\x18\n\x08Operator\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a\x39\n\x07\x46ileLoc\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\x12\x0c\n\x04line\x18\x02 \x01(\r\x12\x0e\n\x06\x63olumn\x18\x03 \x01(\r\x1aU\n\tSourceLoc\x12\x0c\n\x04name\x18\x01 \x01(\t\x12:\n\x06source\x18\x02 \x01(\x0b\x32*.tflite.metrics.ConverterErrorData.FileLoc\x1a\x85\x01\n\x08Location\x12=\n\x04type\x18\x01 \x01(\x0e\x32/.tflite.metrics.ConverterErrorData.LocationType\x12:\n\x04\x63\x61ll\x18\x02 \x03(\x0b\x32,.tflite.metrics.ConverterErrorData.SourceLoc\"\xc5\x01\n\tErrorCode\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x18\n\x14\x45RROR_NEEDS_FLEX_OPS\x10\x01\x12\x1a\n\x16\x45RROR_NEEDS_CUSTOM_OPS\x10\x02\x12%\n!ERROR_UNSUPPORTED_CONTROL_FLOW_V1\x10\x03\x12/\n+ERROR_STATEFUL_PARTITIONED_CALL_IN_FINAL_IR\x10\x04\x12\x1d\n\x18\x45RROR_GPU_NOT_COMPATIBLE\x10\xc8\x01\"J\n\x0cLocationType\x12\x0e\n\nUNKNOWNLOC\x10\x00\x12\x0b\n\x07NAMELOC\x10\x01\x12\x0f\n\x0b\x43\x41LLSITELOC\x10\x02\x12\x0c\n\x08\x46USEDLOC\x10\x03')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tensorflow.compiler.mlir.lite.metrics.converter_error_data_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CONVERTERERRORDATA']._serialized_start=85
  _globals['_CONVERTERERRORDATA']._serialized_end=945
  _globals['_CONVERTERERRORDATA_OPERATOR']._serialized_start=363
  _globals['_CONVERTERERRORDATA_OPERATOR']._serialized_end=387
  _globals['_CONVERTERERRORDATA_FILELOC']._serialized_start=389
  _globals['_CONVERTERERRORDATA_FILELOC']._serialized_end=446
  _globals['_CONVERTERERRORDATA_SOURCELOC']._serialized_start=448
  _globals['_CONVERTERERRORDATA_SOURCELOC']._serialized_end=533
  _globals['_CONVERTERERRORDATA_LOCATION']._serialized_start=536
  _globals['_CONVERTERERRORDATA_LOCATION']._serialized_end=669
  _globals['_CONVERTERERRORDATA_ERRORCODE']._serialized_start=672
  _globals['_CONVERTERERRORDATA_ERRORCODE']._serialized_end=869
  _globals['_CONVERTERERRORDATA_LOCATIONTYPE']._serialized_start=871
  _globals['_CONVERTERERRORDATA_LOCATIONTYPE']._serialized_end=945
# @@protoc_insertion_point(module_scope)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility for working with accelerator systems."""

from typing import List, Optional

from absl import logging

from tensorflow.core.protobuf import cluster_pb2
from tensorflow.core.protobuf import tensorflow_server_pb2
from tensorflow.dtensor.python import config
from tensorflow.dtensor.python import tpu_util
from tensorflow.python.eager import context
from tensorflow.python.framework import config as tf_config
from tensorflow.python.platform import remote_utils
from tensorflow.python.util.tf_export import tf_export

_INITIALIZED_ACCELERATOR_SYSTEM_TYPE = None


def is_initialized() -> bool:
  """Returns whether accelerator system has been initialized."""
  return bool(_INITIALIZED_ACCELERATOR_SYSTEM_TYPE)


def set_initialized(value):
  """Sets if accelerator system has been initialized."""
  global _INITIALIZED_ACCELERATOR_SYSTEM_TYPE
  _INITIALIZED_ACCELERATOR_SYSTEM_TYPE = value


def initialize_multi_client_cluster(job_name: str,
                                    dtensor_jobs: List[str],
                                    client_id: int,
                                    collective_leader: str,
                                    port: Optional[int] = None,
                                    gpu_use_nccl_communication: bool = False,
                                    enable_coordination_service: bool = True):
  """Initialize GRPC servers and collectives for multi-client DTensor setup.

  This function can be used to initialize a multi-client cluster and enable
  collective ops. GRPC servers are necessary in the multi-client mode, even
  when the number of clientis is 1.

  NOTE: this function must be called in an eager context.

  Args:
    job_name: The job name used by all clients in the DTensor cluster.
    dtensor_jobs: A list of the DTensor client jobs participating in the
      cluster. Must be strings of the form "hostname:port".
    client_id: The ID of the DTensor client this function is being called in.
    collective_leader: The job/task that will be used to run collectives.
    port: The port this client's GRPC server will run on. If omitted, use the
      port from dtensor_jobs for this client.
    gpu_use_nccl_communication: if True, configure TensorFlow to use NCCL by
      default.
    enable_coordination_service: If true, enable distributed coordination
      service to make sure that workers know the devices on each other, a
      prerequisite for data transfer through cross-worker rendezvous.

  Raises:
    RuntimeError: If running inside a tf.function.
  """
  assert context.executing_eagerly()

  if not collective_leader.startswith("/job:"):
    collective_leader = "/job:" + collective_leader

  context.context().configure_collective_ops(
      use_nccl_communication=gpu_use_nccl_communication,
      collective_leader=collective_leader)
  if enable_coordination_service:
    context.context().configure_coordination_service(
        service_type="standalone", service_leader=collective_leader)

  config_proto = context.get_config()

  # Construct server def from the host directly instead of relying on
  # TF_CONFIG.
  cluster_def = cluster_pb2.ClusterDef()
  # Note that for bns addresses, we will currently rely on the sorted string
  # of job name as the order of assigning task ids. This might be brittle once
  # we have jobs across multiple cells.
  cluster_def.job.add(name=job_name, tasks=dict(enumerate(dtensor_jobs)))
  server_def = tensorflow_server_pb2.ServerDef(
      cluster=cluster_def,
      default_session_config=config_proto,
      job_name=job_name,
      task_index=client_id,
      protocol=remote_utils.get_default_communication_protocol(),
      port=port)
  server_def.default_session_config.rpc_options.num_channels_per_target = 4
  server_def.default_session_config.experimental.recv_buf_max_chunk = -1

  logging.info("Enabling collectives with server_def: %s", server_def)

  context.context().enable_collective_ops(server_def)

  context.ensure_initialized()


@tf_export(
    "experimental.dtensor.initialize_accelerator_system",
    "experimental.dtensor.initialize_tpu_system",
    "experimental.dtensor.initialize_multi_client",
    v1=[])
def initialize_accelerator_system(
    device_type: Optional[str] = None,
    enable_coordination_service: Optional[bool] = True,
    num_logical_cpu_devices: Optional[int] = None,
    experimental_reset_context: Optional[bool] = False,
    experimental_enable_megcore: Optional[bool] = False,
) -> str:
  """Initializes accelerators and communication fabrics for DTensor.

  DTensor configures TensorFlow to run in the local mode or multi-client mode.
  - In local mode, a mesh can only use devices attached to the current process.
  - In multi-client mode, a mesh can span across devices from multiple clients.

  If `DTENSOR_JOBS` is non-empty, DTensor configures TensorFlow to run in the
  multi-client mode using the distributed runtime. In multi-client mode devices
  on different clients can communicate with each other.

  The following environment variables controls the behavior of this function.

  - `DTENSOR_JOBS`: string, a comma separated list. Each item in the list is
      of format `{hostname}:{port}`. If empty, DTensor runs in the local mode.
      Examples of valid `DTENSOR_JOBS` values:
      - 4 clients on localhost:
        `localhost:10000,localhost:10001,localhost:10002,localhost:10003`
      - 2 clients on host1, 2 clients on host2
        `host1:10000,host1:10001,host2:10000,host2:10003`
      If the hostnames are BNS addresses, the items must be sorted in
      alphabetical order.
  - `DTENSOR_CLIENT_ID`: integer, between `0` to `num_clients - 1`, to identify
      the client id of the current process. The default value is `0`.
  - `DTENSOR_JOB_NAME`: string, a string for the name of the TensorFlow job.
      The job name controls the job name section of the TensorFlow DeviceSpecs,
      e.g., `job:worker` in `/job:worker/replica:0/task:0/device:TPU:0` when
      the job name is `worker`.
      The default value is `localhost` in local mode, and
      `worker` when in the multi-client mode. All DTensor clients within the
      same multi-client cluster share the same job name.

  Args:
    device_type: Type of accelerator to use, can be CPU, GPU, or TPU. If None,
      uses `tf.experimental.dtensor.preferred_device_type()`.
    enable_coordination_service: If true, enable distributed coordination
      service to make sure that workers know the devices on each other, when
      there is more than 1 client.
    num_logical_cpu_devices: the number of logical CPU devices per DTensor
      client. Default to the current number of logical CPU
      (`dtensor.num_local_devices("CPU")`),when `device_type` is CPU, otherwise
      set automatially to match the number of local GPU/TPU devices.
    experimental_reset_context: Reset the tensorflow context. Behaviors of
      existing TensorFlow objects (e.g. Tensors) are undefined. Set this to True
      as an escape hatch, if there is no clear way to refactor your code to call
      initialize_accelerator_system() before calling TensorFlow APIs that
      initialize the context.
    experimental_enable_megcore: Optionally enable megcore in backend.

  Returns:
    device_type: the type of accelerator that was initialized.
  """
  global _INITIALIZED_ACCELERATOR_SYSTEM_TYPE
  assert context.executing_eagerly()

  if is_initialized():
    raise ValueError(
        "Accelerator system has already been initialized. "
        "Call tf.experimental.dtensor.shutdown_accelerator_system() first.")

  if experimental_reset_context:
    if context.context()._initialized:    # pylint: disable=protected-access
      logging.warn(
          "experimental_reset_context is True. "
          "Resetting TensorFlow context. Existing TensorFlow objects "
          "(e.g. Tensors and resources) are invalidated."
      )
      context.context().ensure_uninitialized()

  if context.context()._initialized:  # pylint: disable=protected-access
    raise ValueError(
        "TensorFlow has already been initialized. "
        "tf.experimental.dtensor.initialize_accelerator_system() must be "
        "called before TensorFlow is initialized.")

  context.context()._clear_caches()  # pylint: disable=protected-access

  if device_type is None:
    device_type = config.preferred_device_type()

  device_type = device_type.upper()
  if device_type not in {"CPU", "GPU", "TPU"}:
    raise ValueError(f"Unknown device_type {device_type}. "
                     "Allowed values are CPU, GPU, or TPU")

  if config.gpu_use_nccl_communication():
    logical_gpu_count = config.num_local_devices("GPU")
    physical_gpu_count = len(tf_config.list_physical_devices("GPU"))
    if logical_gpu_count > physical_gpu_count:
      raise ValueError(
          "DTENSOR_GPU_USE_NCCL_COMMUNICATION is set for using NCCL. "
          "NCCL Collectives require one to one mapping between logical and "
          "physical GPUs. "
          f"The number of logical GPU ({logical_gpu_count}) "
          f"is more than the number of physical GPU ({physical_gpu_count})."
      )

  # Configure logical host CPU devices for accelerators.
  if device_type in ("GPU", "TPU"):
    num_local_devices = config.num_local_devices(device_type)
    if num_logical_cpu_devices is None:
      num_logical_cpu_devices = max(
          config.num_local_devices("CPU"), num_local_devices
      )
    else:
      if num_logical_cpu_devices < num_local_devices:
        raise ValueError(
            "If set, `num_logical_cpu_devices`"
            f" (={num_logical_cpu_devices}) must be greater than or"
            f" equal to the number of local {device_type} devices"
            f" (={num_local_devices})"
        )

  if num_logical_cpu_devices is not None:
    tf_config.set_logical_device_configuration(
        tf_config.list_physical_devices("CPU")[0],
        [context.LogicalDeviceConfiguration()]
        * num_logical_cpu_devices,
    )

  if not config.is_local_mode():
    initialize_multi_client_cluster(
        job_name=config.job_name(),
        dtensor_jobs=config.jobs(),
        client_id=config.client_id(),
        collective_leader=config.full_job_name(task_id=0),
        gpu_use_nccl_communication=config.gpu_use_nccl_communication(),
        enable_coordination_service=enable_coordination_service)
  else:
    if device_type == "GPU":
      # Enables Nccl on local mode.
      context.context(  # pylint: disable=protected-access
      )._collective_use_nccl_communication = config.gpu_use_nccl_communication(
      )

  if device_type == "TPU":
    tpu_util.initialize_tpu_system(use_megacore=experimental_enable_megcore)

  _INITIALIZED_ACCELERATOR_SYSTEM_TYPE = device_type

  return device_type


@tf_export(
    "experimental.dtensor.shutdown_accelerator_system",
    "experimental.dtensor.shutdown_tpu_system",
    v1=[])
def shutdown_accelerator_system() -> None:
  """Shuts down the accelerator system."""
  global _INITIALIZED_ACCELERATOR_SYSTEM_TYPE
  try:
    context.async_wait()
  finally:
    if not is_initialized():
      raise ValueError(
          "Accelerator system is not initialized. Call "
          "tf.experimental.dtensor.initialize_accelerator_system first."
      )

    device_type = _INITIALIZED_ACCELERATOR_SYSTEM_TYPE

    if not config.is_local_mode():
      raise ValueError(
          "Shutting down accelerator system under multi-client mode is "
          "not supported."
      )

    if device_type == "TPU":
      tpu_util.shutdown_tpu_system()

    # reset TF context to stop gRPC servers.
    context._reset_context()  # pylint: disable=protected-access
    context.context()._clear_caches()  # pylint: disable=protected-access
    _INITIALIZED_ACCELERATOR_SYSTEM_TYPE = None



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.types.experimental namespace
"""

import sys as _sys

from tensorflow._api.v2.types.experimental import distributed
from tensorflow.python.types.core import AtomicFunction # line: 142
from tensorflow.python.types.core import Callable # line: 111
from tensorflow.python.types.core import ConcreteFunction # line: 169
from tensorflow.python.types.core import FunctionType # line: 90
from tensorflow.python.types.core import PolymorphicFunction as GenericFunction # line: 186
from tensorflow.python.types.core import PolymorphicFunction # line: 186
from tensorflow.python.types.core import TensorLike # line: 417
from tensorflow.python.types.trace import SupportsTracingProtocol # line: 307
from tensorflow.python.types.trace import TraceType # line: 39



# Copyright 2023 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""NoneTensor and NoneTensorSpec classes."""

from tensorflow.python.framework import composite_tensor
from tensorflow.python.framework import type_spec
from tensorflow.python.framework import type_spec_registry


# TODO(b/149584798): add tests for non-tf.data functionality.
class NoneTensor(composite_tensor.CompositeTensor):
  """Composite tensor representation for `None` value."""

  @property
  def _type_spec(self):
    return NoneTensorSpec()


# TODO(b/149584798): add tests for non-tf.data functionality.
@type_spec_registry.register("tf.NoneTensorSpec")
class NoneTensorSpec(type_spec.BatchableTypeSpec):
  """Type specification for `None` value."""

  @property
  def value_type(self):
    return NoneTensor

  def _serialize(self):
    return ()

  @property
  def _component_specs(self):
    return []

  def _to_components(self, value):
    return []

  def _from_components(self, components):
    return

  def _to_tensor_list(self, value):
    return []

  @staticmethod
  def from_value(value):
    return NoneTensorSpec()

  def _batch(self, batch_size):
    return NoneTensorSpec()

  def _unbatch(self):
    return NoneTensorSpec()

  def _to_batched_tensor_list(self, value):
    return []

  def _to_legacy_output_types(self):
    return self

  def _to_legacy_output_shapes(self):
    return self

  def _to_legacy_output_classes(self):
    return self

  def most_specific_compatible_shape(self, other):
    if type(self) is not type(other):
      raise ValueError("No `TypeSpec` is compatible with both {} and {}".format(
          self, other))
    return self


type_spec.register_type_spec_from_value_converter(type(None),
                                                  NoneTensorSpec.from_value)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.signal namespace
"""

import sys as _sys

from tensorflow.python.ops.gen_spectral_ops import fft # line: 353
from tensorflow.python.ops.gen_spectral_ops import fft2d # line: 442
from tensorflow.python.ops.gen_spectral_ops import fft3d # line: 531
from tensorflow.python.ops.gen_spectral_ops import ifft # line: 724
from tensorflow.python.ops.gen_spectral_ops import ifft2d # line: 813
from tensorflow.python.ops.gen_spectral_ops import ifft3d # line: 902
from tensorflow.python.ops.signal.dct_ops import dct # line: 50
from tensorflow.python.ops.signal.dct_ops import idct # line: 212
from tensorflow.python.ops.signal.fft_ops import fftnd # line: 400
from tensorflow.python.ops.signal.fft_ops import fftshift # line: 603
from tensorflow.python.ops.signal.fft_ops import ifftnd # line: 404
from tensorflow.python.ops.signal.fft_ops import ifftshift # line: 648
from tensorflow.python.ops.signal.fft_ops import irfft # line: 411
from tensorflow.python.ops.signal.fft_ops import irfft2d # line: 417
from tensorflow.python.ops.signal.fft_ops import irfft3d # line: 423
from tensorflow.python.ops.signal.fft_ops import irfftnd # line: 430
from tensorflow.python.ops.signal.fft_ops import rfft # line: 408
from tensorflow.python.ops.signal.fft_ops import rfft2d # line: 414
from tensorflow.python.ops.signal.fft_ops import rfft3d # line: 420
from tensorflow.python.ops.signal.fft_ops import rfftnd # line: 426
from tensorflow.python.ops.signal.mel_ops import linear_to_mel_weight_matrix # line: 90
from tensorflow.python.ops.signal.mfcc_ops import mfccs_from_log_mel_spectrograms # line: 25
from tensorflow.python.ops.signal.reconstruction_ops import overlap_and_add # line: 26
from tensorflow.python.ops.signal.shape_ops import frame # line: 55
from tensorflow.python.ops.signal.spectral_ops import inverse_mdct # line: 372
from tensorflow.python.ops.signal.spectral_ops import inverse_stft # line: 160
from tensorflow.python.ops.signal.spectral_ops import inverse_stft_window_fn # line: 95
from tensorflow.python.ops.signal.spectral_ops import mdct # line: 296
from tensorflow.python.ops.signal.spectral_ops import stft # line: 34
from tensorflow.python.ops.signal.window_ops import hamming_window # line: 171
from tensorflow.python.ops.signal.window_ops import hann_window # line: 145
from tensorflow.python.ops.signal.window_ops import kaiser_bessel_derived_window # line: 92
from tensorflow.python.ops.signal.window_ops import kaiser_window # line: 53
from tensorflow.python.ops.signal.window_ops import vorbis_window # line: 120

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "signal", public_apis=None, deprecation=False,
      has_lite=False)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.data namespace
"""

import sys as _sys

from tensorflow._api.v2.compat.v1.data import experimental
from tensorflow.python.data.ops.dataset_ops import AUTOTUNE # line: 103
from tensorflow.python.data.ops.dataset_ops import DatasetV1 as Dataset # line: 3710
from tensorflow.python.data.ops.dataset_ops import DatasetSpec # line: 4594
from tensorflow.python.data.ops.dataset_ops import INFINITE as INFINITE_CARDINALITY # line: 113
from tensorflow.python.data.ops.dataset_ops import NumpyIterator # line: 4759
from tensorflow.python.data.ops.dataset_ops import UNKNOWN as UNKNOWN_CARDINALITY # line: 114
from tensorflow.python.data.ops.dataset_ops import get_legacy_output_classes as get_output_classes # line: 4442
from tensorflow.python.data.ops.dataset_ops import get_legacy_output_shapes as get_output_shapes # line: 4464
from tensorflow.python.data.ops.dataset_ops import get_legacy_output_types as get_output_types # line: 4486
from tensorflow.python.data.ops.dataset_ops import make_initializable_iterator # line: 4353
from tensorflow.python.data.ops.dataset_ops import make_one_shot_iterator # line: 4318
from tensorflow.python.data.ops.iterator_ops import Iterator # line: 95
from tensorflow.python.data.ops.options import Options # line: 579
from tensorflow.python.data.ops.options import ThreadingOptions # line: 535
from tensorflow.python.data.ops.readers import FixedLengthRecordDatasetV1 as FixedLengthRecordDataset # line: 669
from tensorflow.python.data.ops.readers import TFRecordDatasetV1 as TFRecordDataset # line: 490
from tensorflow.python.data.ops.readers import TextLineDatasetV1 as TextLineDataset # line: 260

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "data", public_apis=None, deprecation=False,
      has_lite=False)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.version namespace
"""

import sys as _sys

from tensorflow.python.framework.versions import COMPILER_VERSION # line: 41
from tensorflow.python.framework.versions import GIT_VERSION # line: 35
from tensorflow.python.framework.versions import GRAPH_DEF_VERSION # line: 68
from tensorflow.python.framework.versions import GRAPH_DEF_VERSION_MIN_CONSUMER # line: 74
from tensorflow.python.framework.versions import GRAPH_DEF_VERSION_MIN_PRODUCER # line: 82
from tensorflow.python.framework.versions import VERSION # line: 29

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "version", public_apis=None, deprecation=False,
      has_lite=False)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.nn.rnn_cell namespace
"""

import sys as _sys

from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import BasicLSTMCell # line: 654
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import BasicRNNCell # line: 399
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import DeviceWrapper # line: 1213
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import DropoutWrapper # line: 1191
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import GRUCell # line: 495
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import LSTMCell # line: 824
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import LSTMStateTuple # line: 634
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import MultiRNNCell # line: 1223
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import RNNCell # line: 184
from tensorflow.python.keras.layers.legacy_rnn.rnn_cell_impl import ResidualWrapper # line: 1202

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "nn.rnn_cell", public_apis=None, deprecation=False,
      has_lite=False)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.autograph namespace
"""

import sys as _sys

from tensorflow._api.v2.compat.v2.autograph import experimental
from tensorflow.python.autograph.impl.api import to_code # line: 904
from tensorflow.python.autograph.impl.api import to_graph # line: 707
from tensorflow.python.autograph.utils.ag_logging import set_verbosity # line: 36
from tensorflow.python.autograph.utils.ag_logging import trace # line: 87



# Copyright 2023 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

class Flag:
    def __init__(self, *args, **kwargs) -> None: ...
    def reset(self, arg0: bool) -> None: ...
    def value(self) -> bool: ...

class Flags:
    enable_aggressive_constant_replication: Flag
    enable_colocation_key_propagation_in_while_op_lowering: Flag
    enable_function_pruning_before_inlining: Flag
    enable_graph_debug_info_caching_for_stack_frames: Flag
    enable_nested_function_shape_inference: Flag
    enable_quantized_dtypes_training: Flag
    enable_skip_encapsulation_for_non_tpu_graphs: Flag
    enable_tf2min_ici_weight: Flag
    graph_building_optimization: Flag
    more_stack_traces: Flag
    op_building_optimization: Flag
    publish_function_graphs: Flag
    saved_model_fingerprinting: Flag
    test_only_experiment_1: Flag
    test_only_experiment_2: Flag
    tf_shape_default_int64: Flag
    def __init__(self) -> None: ...



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.random.experimental namespace
"""

import sys as _sys

from tensorflow.python.ops.random_ops_util import Algorithm # line: 30
from tensorflow.python.ops.stateful_random_ops import Generator # line: 220
from tensorflow.python.ops.stateful_random_ops import create_rng_state # line: 174
from tensorflow.python.ops.stateful_random_ops import get_global_generator # line: 975
from tensorflow.python.ops.stateful_random_ops import set_global_generator # line: 1000
from tensorflow.python.ops.stateless_random_ops import index_shuffle # line: 135
from tensorflow.python.ops.stateless_random_ops import fold_in as stateless_fold_in # line: 93
from tensorflow.python.ops.stateless_random_ops import stateless_shuffle # line: 226
from tensorflow.python.ops.stateless_random_ops import split as stateless_split # line: 51



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The implementation of `tf.data.Dataset.batch`."""

import warnings

from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.data.ops import debug_mode
from tensorflow.python.data.util import nest
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_util
from tensorflow.python.ops import gen_dataset_ops


def _batch(input_dataset,
           batch_size,
           drop_remainder=False,
           num_parallel_calls=None,
           deterministic=None,
           name=None):
  """See `Dataset.batch` for details."""
  if num_parallel_calls is None or debug_mode.DEBUG_MODE:
    if deterministic is not None and not debug_mode.DEBUG_MODE:
      warnings.warn("The `deterministic` argument has no effect unless the "
                    "`num_parallel_calls` argument is specified.")
    return _BatchDataset(input_dataset, batch_size, drop_remainder, name=name)
  else:
    return _ParallelBatchDataset(
        input_dataset,
        batch_size,
        drop_remainder,
        num_parallel_calls,
        deterministic,
        name=name)


class _BatchDataset(dataset_ops.UnaryDataset):
  """A `Dataset` that batches contiguous elements from its input."""

  def __init__(self, input_dataset, batch_size, drop_remainder, name=None):
    """See `Dataset.batch()` for details."""
    self._input_dataset = input_dataset
    self._batch_size = ops.convert_to_tensor(
        batch_size, dtype=dtypes.int64, name="batch_size")
    self._drop_remainder = ops.convert_to_tensor(
        drop_remainder, dtype=dtypes.bool, name="drop_remainder")

    constant_drop_remainder = tensor_util.constant_value(self._drop_remainder)
    # pylint: disable=protected-access
    if constant_drop_remainder:
      # NOTE(mrry): `constant_drop_remainder` may be `None` (unknown statically)
      # or `False` (explicitly retaining the remainder).
      # pylint: disable=g-long-lambda
      constant_batch_size = tensor_util.constant_value(self._batch_size)
      self._structure = nest.map_structure(
          lambda component_spec: component_spec._batch(constant_batch_size),
          input_dataset.element_spec)
    else:
      self._structure = nest.map_structure(
          lambda component_spec: component_spec._batch(None),
          input_dataset.element_spec)

    self._name = name
    variant_tensor = gen_dataset_ops.batch_dataset_v2(
        input_dataset._variant_tensor,
        batch_size=self._batch_size,
        drop_remainder=self._drop_remainder,
        **self._common_args)
    super().__init__(input_dataset, variant_tensor)

  @property
  def element_spec(self):
    return self._structure


class _ParallelBatchDataset(dataset_ops.UnaryDataset):
  """A `Dataset` that batches contiguous elements from its input in parallel."""

  def __init__(self,
               input_dataset,
               batch_size,
               drop_remainder,
               num_parallel_calls,
               deterministic,
               name=None):
    """See `Dataset.batch()` for details."""
    self._input_dataset = input_dataset
    self._batch_size = ops.convert_to_tensor(
        batch_size, dtype=dtypes.int64, name="batch_size")
    self._drop_remainder = ops.convert_to_tensor(
        drop_remainder, dtype=dtypes.bool, name="drop_remainder")
    self._num_parallel_calls = ops.convert_to_tensor(
        num_parallel_calls, dtype=dtypes.int64, name="num_parallel_calls")
    if deterministic is None:
      self._deterministic = "default"
    elif deterministic:
      self._deterministic = "true"
    else:
      self._deterministic = "false"

    constant_drop_remainder = tensor_util.constant_value(self._drop_remainder)
    # pylint: disable=protected-access
    if constant_drop_remainder:
      # NOTE(mrry): `constant_drop_remainder` may be `None` (unknown statically)
      # or `False` (explicitly retaining the remainder).
      # pylint: disable=g-long-lambda
      constant_batch_size = tensor_util.constant_value(self._batch_size)
      self._structure = nest.map_structure(
          lambda component_spec: component_spec._batch(constant_batch_size),
          input_dataset.element_spec)
    else:
      self._structure = nest.map_structure(
          lambda component_spec: component_spec._batch(None),
          input_dataset.element_spec)

    self._name = name
    variant_tensor = gen_dataset_ops.parallel_batch_dataset(
        input_dataset._variant_tensor,
        batch_size=self._batch_size,
        num_parallel_calls=self._num_parallel_calls,
        drop_remainder=self._drop_remainder,
        deterministic=self._deterministic,
        **self._common_args)

    super().__init__(input_dataset, variant_tensor)

  @property
  def element_spec(self):
    return self._structure



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.debugging namespace
"""

import sys as _sys

from tensorflow._api.v2.compat.v2.debugging import experimental
from tensorflow.python.ops.gen_array_ops import check_numerics # line: 950
from tensorflow.python.debug.lib.check_numerics_callback import disable_check_numerics # line: 445
from tensorflow.python.debug.lib.check_numerics_callback import enable_check_numerics # line: 336
from tensorflow.python.eager.context import get_log_device_placement # line: 2802
from tensorflow.python.eager.context import set_log_device_placement # line: 2812
from tensorflow.python.ops.check_ops import assert_equal_v2 as assert_equal # line: 762
from tensorflow.python.ops.check_ops import assert_greater_v2 as assert_greater # line: 978
from tensorflow.python.ops.check_ops import assert_greater_equal_v2 as assert_greater_equal # line: 996
from tensorflow.python.ops.check_ops import assert_integer_v2 as assert_integer # line: 1434
from tensorflow.python.ops.check_ops import assert_less_v2 as assert_less # line: 942
from tensorflow.python.ops.check_ops import assert_less_equal_v2 as assert_less_equal # line: 959
from tensorflow.python.ops.check_ops import assert_near_v2 as assert_near # line: 803
from tensorflow.python.ops.check_ops import assert_negative_v2 as assert_negative # line: 543
from tensorflow.python.ops.check_ops import assert_non_negative_v2 as assert_non_negative # line: 650
from tensorflow.python.ops.check_ops import assert_non_positive_v2 as assert_non_positive # line: 706
from tensorflow.python.ops.check_ops import assert_none_equal_v2 as assert_none_equal # line: 783
from tensorflow.python.ops.check_ops import assert_positive_v2 as assert_positive # line: 597
from tensorflow.python.ops.check_ops import assert_proper_iterable # line: 511
from tensorflow.python.ops.check_ops import assert_rank_v2 as assert_rank # line: 1064
from tensorflow.python.ops.check_ops import assert_rank_at_least_v2 as assert_rank_at_least # line: 1162
from tensorflow.python.ops.check_ops import assert_rank_in_v2 as assert_rank_in # line: 1329
from tensorflow.python.ops.check_ops import assert_same_float_dtype # line: 2126
from tensorflow.python.ops.check_ops import assert_scalar_v2 as assert_scalar # line: 2161
from tensorflow.python.ops.check_ops import assert_shapes_v2 as assert_shapes # line: 1618
from tensorflow.python.ops.check_ops import assert_type_v2 as assert_type # line: 1494
from tensorflow.python.ops.check_ops import is_numeric_tensor # line: 1961
from tensorflow.python.ops.control_flow_assert import Assert # line: 62
from tensorflow.python.ops.numerics import verify_tensor_all_finite_v2 as assert_all_finite # line: 50
from tensorflow.python.util.traceback_utils import disable_traceback_filtering # line: 76
from tensorflow.python.util.traceback_utils import enable_traceback_filtering # line: 51
from tensorflow.python.util.traceback_utils import is_traceback_filtering_enabled # line: 32



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implementation of LoadDataset in Python."""
import multiprocessing
import os
import time
from typing import Any, Callable, Optional, Union

from absl import logging

from google.protobuf import message
from google.protobuf import text_format
from tensorflow.core.protobuf import snapshot_pb2
from tensorflow.python.data.experimental.service import _pywrap_snapshot_utils
from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.data.ops import structured_function
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import errors
from tensorflow.python.framework import tensor_spec
from tensorflow.python.ops import gen_experimental_dataset_ops as ged_ops
from tensorflow.python.platform import gfile
# TODO(b/238903802): Use TypeSpec serialization methods directly.
from tensorflow.python.saved_model import nested_structure_coder

# For distributed snapshot load V2, retries loading after this time, if the
# snapshot is not ready yet.
_RETRY_INTERVAL_SEC = 5


def _load(  # pylint: disable=unused-private-name
    path: str,
    element_spec: Any,
    compression: Optional[str],
    reader_func: Optional[Callable[[dataset_ops.Dataset], dataset_ops.Dataset]],
    wait: bool,
) -> dataset_ops.Dataset:
  """Loads dataset from tf.data snapshot."""

  if wait:
    return _load_with_retry(path, element_spec, compression, reader_func)

  if reader_func is None:
    reader_func = lambda datasets: datasets.interleave(  # pylint:disable=g-long-lambda
        lambda x: x,
        cycle_length=multiprocessing.cpu_count(),
        num_parallel_calls=dataset_ops.AUTOTUNE)

  distributed_snapshot_metadata = _load_distributed_snapshot_metadata(path)
  if distributed_snapshot_metadata:
    _validate_snapshot(
        path, distributed_snapshot_metadata, element_spec, compression)
    return _load_distributed_snapshot(
        path, distributed_snapshot_metadata, reader_func)

  if element_spec is None:
    element_spec = _load_element_spec(path)
  return _LoadDataset(path, element_spec, compression, reader_func)


def _load_with_retry(  # pylint: disable=unused-private-name
    path: str,
    element_spec: Any = None,
    compression: Optional[str] = None,
    reader_func: Optional[
        Callable[[dataset_ops.Dataset], dataset_ops.Dataset]] = None,
) -> dataset_ops.Dataset:
  """Tries loading the snapshot. Retries if not found."""

  while True:
    try:
      dataset = dataset_ops.Dataset.load(
          path=path,
          element_spec=element_spec,
          compression=compression,
          reader_func=reader_func,
          wait=False)
      logging.info("Load tf.data snapshot at %s.", path)
      return dataset
    except (errors.NotFoundError, FileNotFoundError):
      logging.info(
          "Could not find tf.data snapshot at %s. Will wait and retry.", path)
      time.sleep(_RETRY_INTERVAL_SEC)


def _load_distributed_snapshot_metadata(
    path: str,
) -> Optional[snapshot_pb2.DistributedSnapshotMetadata]:
  """Reads the distributed snapshot metadata.

  Args:
    path: Base path of the snapshot.

  Returns:
    DistributedSnapshotMetadata if the snapshot is a distributed snapshot.
    Returns None if it is a non-distributed snapshot.
  """
  metadata_file = _pywrap_snapshot_utils.TF_DATA_SnapshotMetadataFilePath(path)
  if not gfile.Exists(metadata_file):
    return None

  try:
    with gfile.GFile(metadata_file, "r") as f:
      return text_format.ParseLines(
          f, snapshot_pb2.DistributedSnapshotMetadata())
  except (
      errors.NotFoundError,
      text_format.ParseError,
      message.DecodeError,
      UnicodeDecodeError):
    return None


def _load_distributed_snapshot(
    path: str,
    metadata: snapshot_pb2.DistributedSnapshotMetadata,
    reader_func: Callable[[dataset_ops.Dataset], dataset_ops.Dataset],
) -> dataset_ops.Dataset:
  """Loads a distributed snapshot."""

  dataset = _ListSnapshotChunksDataset(path)
  dataset = dataset.map(
      lambda chunk_file: _SnapshotChunkDataset(  # pylint:disable=g-long-lambda
          chunk_file,
          element_spec=_parse_element_spec(metadata.element_spec),
          compression=metadata.compression))
  return reader_func(dataset)


def _load_element_spec(path: str) -> Any:
  """Loads the dataset element spec.

  Args:
    path: Base path of the snapshot.

  Returns:
    Dataset element_spec.

  Raises:
    NotFoundError if the element spec file does not exist or cannot be decoded.
  """
  dataset_spec_filename = os.path.join(path, dataset_ops.DATASET_SPEC_FILENAME)
  if not gfile.Exists(dataset_spec_filename):
    raise errors.NotFoundError(
        node_def=None, op=None,
        message="tf.data snapshot element_spec file not found: "
                f"{dataset_spec_filename}.")

  with gfile.GFile(dataset_spec_filename, "rb") as f:
    encoded_spec = f.read()
  try:
    return _parse_element_spec(encoded_spec)
  except nested_structure_coder.NotEncodableError as e:
    raise errors.NotFoundError(
        node_def=None, op=None,
        message="tf.data snapshot element_spec file not found or invalid: "
                f"{dataset_spec_filename}.") from e


def _parse_element_spec(encoded_element_spec: Union[bytes, str]) -> Any:
  struct_pb = nested_structure_coder.struct_pb2.StructuredValue()
  struct_pb.ParseFromString(encoded_element_spec)
  return nested_structure_coder.decode_proto(struct_pb)


class _LoadDataset(dataset_ops.DatasetSource):
  """A dataset that loads previously saved dataset."""

  def __init__(
      self,
      path: str,
      element_spec: Any,
      compression: str,
      reader_func: Callable[[dataset_ops.Dataset], dataset_ops.Dataset]):
    self._path = path
    self._element_spec = element_spec
    self._compression = compression
    self._reader_func = structured_function.StructuredFunctionWrapper(
        reader_func,
        "load()",
        # Dataset of datasets of input elements
        input_structure=dataset_ops.DatasetSpec(
            dataset_ops.DatasetSpec(self._element_spec)))

    variant_tensor = ged_ops.load_dataset(
        path,
        reader_func_other_args=self._reader_func.function.captured_inputs,
        compression=compression,
        reader_func=self._reader_func.function,
        **self._flat_structure)
    super().__init__(variant_tensor)

  @property
  def element_spec(self) -> Any:
    return self._element_spec


class _SnapshotChunkDataset(dataset_ops.DatasetSource):
  """A dataset for one chunk file from a tf.data distributed snapshot."""

  def __init__(self, chunk_file: str, element_spec: Any, compression: str):
    self._chunk_file = chunk_file
    self._element_spec = element_spec
    variant_tensor = ged_ops.snapshot_chunk_dataset(
        chunk_file,
        compression=compression,
        **self._flat_structure)
    super().__init__(variant_tensor)

  @property
  def element_spec(self) -> Any:
    return self._element_spec


class _ListSnapshotChunksDataset(dataset_ops.DatasetSource):
  """A dataset for listing snapshot chunk files.

  It supports listing partially written snapshots. When a snapshot is being
  written, it returns the currently available chunk files.
  """

  def __init__(self, snapshot_path: str):
    self._snapshot_path = snapshot_path
    variant_tensor = ged_ops.list_snapshot_chunks_dataset(
        snapshot_path, **self._flat_structure)
    super().__init__(variant_tensor)

  @property
  def element_spec(self) -> tensor_spec.TensorSpec:
    return tensor_spec.TensorSpec([], dtypes.string)


def _validate_snapshot(
    path: str,
    metadata: snapshot_pb2.DistributedSnapshotMetadata,
    element_spec: Any,
    compression: str) -> None:
  """Validates a tf.data distributed snapshot.

  Args:
    path: Root path of the distributed snapshot.
    metadata: The DistributedSnapshotMetadata of the snapshot.
    element_spec: Dataset element_spec.
    compression: Compression method used for saving.

  Raises:
    ValueError if the snapshot is invalid.
  """

  error_file = _pywrap_snapshot_utils.TF_DATA_SnapshotErrorFilePath(path)
  if gfile.Exists(error_file):
    with gfile.GFile(error_file, "r") as f:
      raise ValueError(
          f"Failed to load tf.data snapshot at {path}. The save job failed to "
          f"write it. Status: {f.read()}")

  snapshot_element_spec = _parse_element_spec(metadata.element_spec)
  if element_spec and element_spec != snapshot_element_spec:
    raise ValueError(
        f"Failed to load tf.data snapshot at {path}. User specified "
        f"element_spec {element_spec}, but the actual element_spec is "
        f"{snapshot_element_spec}.")

  if compression and compression != metadata.compression:
    raise ValueError(
        f"Failed to load tf.data snapshot at {path}. User specified "
        f"compression {compression}, but the actual compression is "
        f"{metadata.compression}.")



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.distribute.experimental namespace
"""

import sys as _sys

from tensorflow.python.distribute.central_storage_strategy import CentralStorageStrategyV1 as CentralStorageStrategy # line: 212
from tensorflow.python.distribute.collective_all_reduce_strategy import CollectiveAllReduceStrategyV1 as MultiWorkerMirroredStrategy # line: 290
from tensorflow.python.distribute.collective_util import CommunicationImplementation as CollectiveCommunication # line: 26
from tensorflow.python.distribute.collective_util import Hints as CollectiveHints # line: 168
from tensorflow.python.distribute.collective_util import CommunicationImplementation # line: 26
from tensorflow.python.distribute.collective_util import _OptionsExported as CommunicationOptions # line: 50
from tensorflow.python.distribute.parameter_server_strategy import ParameterServerStrategyV1 as ParameterServerStrategy # line: 48
from tensorflow.python.distribute.tpu_strategy import TPUStrategyV1 as TPUStrategy # line: 752

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "distribute.experimental", public_apis=None, deprecation=False,
      has_lite=False)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.experimental.numpy namespace
"""

import sys as _sys

from tensorflow._api.v2.experimental.numpy import random
from tensorflow.python.framework.tensor import Tensor as ndarray # line: 138
from tensorflow.python.ops.numpy_ops.np_array_ops import all # line: 420
from tensorflow.python.ops.numpy_ops.np_array_ops import amax # line: 672
from tensorflow.python.ops.numpy_ops.np_array_ops import amin # line: 689
from tensorflow.python.ops.numpy_ops.np_array_ops import any # line: 427
from tensorflow.python.ops.numpy_ops.np_array_ops import arange # line: 259
from tensorflow.python.ops.numpy_ops.np_array_ops import around # line: 820
from tensorflow.python.ops.numpy_ops.np_array_ops import array # line: 217
from tensorflow.python.ops.numpy_ops.np_array_ops import asanyarray # line: 246
from tensorflow.python.ops.numpy_ops.np_array_ops import asarray # line: 234
from tensorflow.python.ops.numpy_ops.np_array_ops import ascontiguousarray # line: 252
from tensorflow.python.ops.numpy_ops.np_array_ops import atleast_1d # line: 1293
from tensorflow.python.ops.numpy_ops.np_array_ops import atleast_2d # line: 1299
from tensorflow.python.ops.numpy_ops.np_array_ops import atleast_3d # line: 1305
from tensorflow.python.ops.numpy_ops.np_array_ops import broadcast_arrays # line: 1567
from tensorflow.python.ops.numpy_ops.np_array_ops import broadcast_to # line: 1192
from tensorflow.python.ops.numpy_ops.np_array_ops import compress # line: 434
from tensorflow.python.ops.numpy_ops.np_array_ops import copy # line: 465
from tensorflow.python.ops.numpy_ops.np_array_ops import cumprod # line: 483
from tensorflow.python.ops.numpy_ops.np_array_ops import cumsum # line: 500
from tensorflow.python.ops.numpy_ops.np_array_ops import diag # line: 305
from tensorflow.python.ops.numpy_ops.np_array_ops import diag_indices # line: 1338
from tensorflow.python.ops.numpy_ops.np_array_ops import diagflat # line: 389
from tensorflow.python.ops.numpy_ops.np_array_ops import diagonal # line: 347
from tensorflow.python.ops.numpy_ops.np_array_ops import dsplit # line: 1187
from tensorflow.python.ops.numpy_ops.np_array_ops import dstack # line: 1241
from tensorflow.python.ops.numpy_ops.np_array_ops import empty # line: 53
from tensorflow.python.ops.numpy_ops.np_array_ops import empty_like # line: 59
from tensorflow.python.ops.numpy_ops.np_array_ops import expand_dims # line: 875
from tensorflow.python.ops.numpy_ops.np_array_ops import eye # line: 98
from tensorflow.python.ops.numpy_ops.np_array_ops import flatten # line: 889
from tensorflow.python.ops.numpy_ops.np_array_ops import flip # line: 1433
from tensorflow.python.ops.numpy_ops.np_array_ops import fliplr # line: 1452
from tensorflow.python.ops.numpy_ops.np_array_ops import flipud # line: 1446
from tensorflow.python.ops.numpy_ops.np_array_ops import full # line: 136
from tensorflow.python.ops.numpy_ops.np_array_ops import full_like # line: 148
from tensorflow.python.ops.numpy_ops.np_array_ops import hsplit # line: 1184
from tensorflow.python.ops.numpy_ops.np_array_ops import hstack # line: 1214
from tensorflow.python.ops.numpy_ops.np_array_ops import identity # line: 130
from tensorflow.python.ops.numpy_ops.np_array_ops import imag # line: 517
from tensorflow.python.ops.numpy_ops.np_array_ops import isscalar # line: 1114
from tensorflow.python.ops.numpy_ops.np_array_ops import ix_ # line: 1534
from tensorflow.python.ops.numpy_ops.np_array_ops import max # line: 1672
from tensorflow.python.ops.numpy_ops.np_array_ops import mean # line: 657
from tensorflow.python.ops.numpy_ops.np_array_ops import min # line: 1678
from tensorflow.python.ops.numpy_ops.np_array_ops import moveaxis # line: 953
from tensorflow.python.ops.numpy_ops.np_array_ops import ndim # line: 1107
from tensorflow.python.ops.numpy_ops.np_array_ops import newaxis # line: 48
from tensorflow.python.ops.numpy_ops.np_array_ops import nonzero # line: 1324
from tensorflow.python.ops.numpy_ops.np_array_ops import ones # line: 83
from tensorflow.python.ops.numpy_ops.np_array_ops import ones_like # line: 91
from tensorflow.python.ops.numpy_ops.np_array_ops import pad # line: 1018
from tensorflow.python.ops.numpy_ops.np_array_ops import prod # line: 644
from tensorflow.python.ops.numpy_ops.np_array_ops import ravel # line: 772
from tensorflow.python.ops.numpy_ops.np_array_ops import real # line: 779
from tensorflow.python.ops.numpy_ops.np_array_ops import repeat # line: 788
from tensorflow.python.ops.numpy_ops.np_array_ops import reshape # line: 843
from tensorflow.python.ops.numpy_ops.np_array_ops import roll # line: 1458
from tensorflow.python.ops.numpy_ops.np_array_ops import rot90 # line: 1472
from tensorflow.python.ops.numpy_ops.np_array_ops import round # line: 1684
from tensorflow.python.ops.numpy_ops.np_array_ops import select # line: 1077
from tensorflow.python.ops.numpy_ops.np_array_ops import shape # line: 1095
from tensorflow.python.ops.numpy_ops.np_array_ops import sign # line: 1580
from tensorflow.python.ops.numpy_ops.np_array_ops import size # line: 615
from tensorflow.python.ops.numpy_ops.np_array_ops import split # line: 1152
from tensorflow.python.ops.numpy_ops.np_array_ops import squeeze # line: 882
from tensorflow.python.ops.numpy_ops.np_array_ops import stack # line: 1198
from tensorflow.python.ops.numpy_ops.np_array_ops import std # line: 759
from tensorflow.python.ops.numpy_ops.np_array_ops import sum # line: 631
from tensorflow.python.ops.numpy_ops.np_array_ops import swapaxes # line: 914
from tensorflow.python.ops.numpy_ops.np_array_ops import take # line: 1036
from tensorflow.python.ops.numpy_ops.np_array_ops import take_along_axis # line: 1607
from tensorflow.python.ops.numpy_ops.np_array_ops import transpose # line: 905
from tensorflow.python.ops.numpy_ops.np_array_ops import tri # line: 1353
from tensorflow.python.ops.numpy_ops.np_array_ops import tril # line: 1383
from tensorflow.python.ops.numpy_ops.np_array_ops import triu # line: 1408
from tensorflow.python.ops.numpy_ops.np_array_ops import vander # line: 1493
from tensorflow.python.ops.numpy_ops.np_array_ops import var # line: 706
from tensorflow.python.ops.numpy_ops.np_array_ops import vsplit # line: 1181
from tensorflow.python.ops.numpy_ops.np_array_ops import vstack # line: 1230
from tensorflow.python.ops.numpy_ops.np_array_ops import where # line: 1064
from tensorflow.python.ops.numpy_ops.np_array_ops import zeros # line: 65
from tensorflow.python.ops.numpy_ops.np_array_ops import zeros_like # line: 74
from tensorflow.python.ops.numpy_ops.np_config import enable_numpy_behavior as experimental_enable_numpy_behavior # line: 25
from tensorflow.python.ops.numpy_ops.np_dtypes import bool_ # line: 27
from tensorflow.python.ops.numpy_ops.np_dtypes import complex128 # line: 31
from tensorflow.python.ops.numpy_ops.np_dtypes import complex64 # line: 35
from tensorflow.python.ops.numpy_ops.np_dtypes import complex_ # line: 112
from tensorflow.python.ops.numpy_ops.np_dtypes import float16 # line: 39
from tensorflow.python.ops.numpy_ops.np_dtypes import float32 # line: 43
from tensorflow.python.ops.numpy_ops.np_dtypes import float64 # line: 47
from tensorflow.python.ops.numpy_ops.np_dtypes import float_ # line: 115
from tensorflow.python.ops.numpy_ops.np_dtypes import iinfo # line: 121
from tensorflow.python.ops.numpy_ops.np_dtypes import inexact # line: 51
from tensorflow.python.ops.numpy_ops.np_dtypes import int16 # line: 59
from tensorflow.python.ops.numpy_ops.np_dtypes import int32 # line: 63
from tensorflow.python.ops.numpy_ops.np_dtypes import int64 # line: 67
from tensorflow.python.ops.numpy_ops.np_dtypes import int8 # line: 71
from tensorflow.python.ops.numpy_ops.np_dtypes import int_ # line: 55
from tensorflow.python.ops.numpy_ops.np_dtypes import issubdtype # line: 126
from tensorflow.python.ops.numpy_ops.np_dtypes import object_ # line: 75
from tensorflow.python.ops.numpy_ops.np_dtypes import string_ # line: 80
from tensorflow.python.ops.numpy_ops.np_dtypes import uint16 # line: 84
from tensorflow.python.ops.numpy_ops.np_dtypes import uint32 # line: 88
from tensorflow.python.ops.numpy_ops.np_dtypes import uint64 # line: 92
from tensorflow.python.ops.numpy_ops.np_dtypes import uint8 # line: 96
from tensorflow.python.ops.numpy_ops.np_dtypes import unicode_ # line: 101
from tensorflow.python.ops.numpy_ops.np_math_ops import abs # line: 698
from tensorflow.python.ops.numpy_ops.np_math_ops import absolute # line: 704
from tensorflow.python.ops.numpy_ops.np_math_ops import add # line: 90
from tensorflow.python.ops.numpy_ops.np_math_ops import allclose # line: 542
from tensorflow.python.ops.numpy_ops.np_math_ops import angle # line: 852
from tensorflow.python.ops.numpy_ops.np_math_ops import append # line: 1427
from tensorflow.python.ops.numpy_ops.np_math_ops import arccos # line: 799
from tensorflow.python.ops.numpy_ops.np_math_ops import arccosh # line: 817
from tensorflow.python.ops.numpy_ops.np_math_ops import arcsin # line: 793
from tensorflow.python.ops.numpy_ops.np_math_ops import arcsinh # line: 811
from tensorflow.python.ops.numpy_ops.np_math_ops import arctan # line: 805
from tensorflow.python.ops.numpy_ops.np_math_ops import arctan2 # line: 402
from tensorflow.python.ops.numpy_ops.np_math_ops import arctanh # line: 823
from tensorflow.python.ops.numpy_ops.np_math_ops import argmax # line: 1415
from tensorflow.python.ops.numpy_ops.np_math_ops import argmin # line: 1421
from tensorflow.python.ops.numpy_ops.np_math_ops import argsort # line: 1350
from tensorflow.python.ops.numpy_ops.np_math_ops import array_equal # line: 1155
from tensorflow.python.ops.numpy_ops.np_math_ops import average # line: 1436
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_and # line: 630
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_not # line: 648
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_or # line: 636
from tensorflow.python.ops.numpy_ops.np_math_ops import bitwise_xor # line: 642
from tensorflow.python.ops.numpy_ops.np_math_ops import cbrt # line: 868
from tensorflow.python.ops.numpy_ops.np_math_ops import ceil # line: 716
from tensorflow.python.ops.numpy_ops.np_math_ops import clip # line: 222
from tensorflow.python.ops.numpy_ops.np_math_ops import concatenate # line: 1308
from tensorflow.python.ops.numpy_ops.np_math_ops import conj # line: 728
from tensorflow.python.ops.numpy_ops.np_math_ops import conjugate # line: 879
from tensorflow.python.ops.numpy_ops.np_math_ops import cos # line: 763
from tensorflow.python.ops.numpy_ops.np_math_ops import cosh # line: 781
from tensorflow.python.ops.numpy_ops.np_math_ops import count_nonzero # line: 1344
from tensorflow.python.ops.numpy_ops.np_math_ops import cross # line: 289
from tensorflow.python.ops.numpy_ops.np_math_ops import deg2rad # line: 829
from tensorflow.python.ops.numpy_ops.np_math_ops import diff # line: 1054
from tensorflow.python.ops.numpy_ops.np_math_ops import divide # line: 144
from tensorflow.python.ops.numpy_ops.np_math_ops import divmod # line: 182
from tensorflow.python.ops.numpy_ops.np_math_ops import dot # line: 60
from tensorflow.python.ops.numpy_ops.np_math_ops import e # line: 51
from tensorflow.python.ops.numpy_ops.np_math_ops import einsum # line: 1553
from tensorflow.python.ops.numpy_ops.np_math_ops import equal # line: 1119
from tensorflow.python.ops.numpy_ops.np_math_ops import exp # line: 686
from tensorflow.python.ops.numpy_ops.np_math_ops import exp2 # line: 885
from tensorflow.python.ops.numpy_ops.np_math_ops import expm1 # line: 894
from tensorflow.python.ops.numpy_ops.np_math_ops import fabs # line: 710
from tensorflow.python.ops.numpy_ops.np_math_ops import fix # line: 900
from tensorflow.python.ops.numpy_ops.np_math_ops import float_power # line: 396
from tensorflow.python.ops.numpy_ops.np_math_ops import floor # line: 722
from tensorflow.python.ops.numpy_ops.np_math_ops import floor_divide # line: 150
from tensorflow.python.ops.numpy_ops.np_math_ops import gcd # line: 588
from tensorflow.python.ops.numpy_ops.np_math_ops import geomspace # line: 1269
from tensorflow.python.ops.numpy_ops.np_math_ops import greater # line: 1131
from tensorflow.python.ops.numpy_ops.np_math_ops import greater_equal # line: 1137
from tensorflow.python.ops.numpy_ops.np_math_ops import heaviside # line: 414
from tensorflow.python.ops.numpy_ops.np_math_ops import hypot # line: 430
from tensorflow.python.ops.numpy_ops.np_math_ops import inf # line: 55
from tensorflow.python.ops.numpy_ops.np_math_ops import inner # line: 273
from tensorflow.python.ops.numpy_ops.np_math_ops import isclose # line: 524
from tensorflow.python.ops.numpy_ops.np_math_ops import iscomplex # line: 909
from tensorflow.python.ops.numpy_ops.np_math_ops import iscomplexobj # line: 921
from tensorflow.python.ops.numpy_ops.np_math_ops import isfinite # line: 982
from tensorflow.python.ops.numpy_ops.np_math_ops import isinf # line: 988
from tensorflow.python.ops.numpy_ops.np_math_ops import isnan # line: 934
from tensorflow.python.ops.numpy_ops.np_math_ops import isneginf # line: 996
from tensorflow.python.ops.numpy_ops.np_math_ops import isposinf # line: 1004
from tensorflow.python.ops.numpy_ops.np_math_ops import isreal # line: 915
from tensorflow.python.ops.numpy_ops.np_math_ops import isrealobj # line: 928
from tensorflow.python.ops.numpy_ops.np_math_ops import kron # line: 436
from tensorflow.python.ops.numpy_ops.np_math_ops import lcm # line: 595
from tensorflow.python.ops.numpy_ops.np_math_ops import less # line: 1143
from tensorflow.python.ops.numpy_ops.np_math_ops import less_equal # line: 1149
from tensorflow.python.ops.numpy_ops.np_math_ops import linspace # line: 1205
from tensorflow.python.ops.numpy_ops.np_math_ops import log # line: 680
from tensorflow.python.ops.numpy_ops.np_math_ops import log10 # line: 1018
from tensorflow.python.ops.numpy_ops.np_math_ops import log1p # line: 1024
from tensorflow.python.ops.numpy_ops.np_math_ops import log2 # line: 1012
from tensorflow.python.ops.numpy_ops.np_math_ops import logaddexp # line: 482
from tensorflow.python.ops.numpy_ops.np_math_ops import logaddexp2 # line: 494
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_and # line: 1180
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_not # line: 1198
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_or # line: 1186
from tensorflow.python.ops.numpy_ops.np_math_ops import logical_xor # line: 1192
from tensorflow.python.ops.numpy_ops.np_math_ops import logspace # line: 1251
from tensorflow.python.ops.numpy_ops.np_math_ops import matmul # line: 236
from tensorflow.python.ops.numpy_ops.np_math_ops import maximum # line: 188
from tensorflow.python.ops.numpy_ops.np_math_ops import meshgrid # line: 1524
from tensorflow.python.ops.numpy_ops.np_math_ops import minimum # line: 210
from tensorflow.python.ops.numpy_ops.np_math_ops import mod # line: 163
from tensorflow.python.ops.numpy_ops.np_math_ops import multiply # line: 108
from tensorflow.python.ops.numpy_ops.np_math_ops import nanmean # line: 965
from tensorflow.python.ops.numpy_ops.np_math_ops import nanprod # line: 960
from tensorflow.python.ops.numpy_ops.np_math_ops import nansum # line: 957
from tensorflow.python.ops.numpy_ops.np_math_ops import negative # line: 734
from tensorflow.python.ops.numpy_ops.np_math_ops import nextafter # line: 408
from tensorflow.python.ops.numpy_ops.np_math_ops import not_equal # line: 1125
from tensorflow.python.ops.numpy_ops.np_math_ops import outer # line: 472
from tensorflow.python.ops.numpy_ops.np_math_ops import pi # line: 47
from tensorflow.python.ops.numpy_ops.np_math_ops import polyval # line: 506
from tensorflow.python.ops.numpy_ops.np_math_ops import positive # line: 1030
from tensorflow.python.ops.numpy_ops.np_math_ops import power # line: 390
from tensorflow.python.ops.numpy_ops.np_math_ops import ptp # line: 1300
from tensorflow.python.ops.numpy_ops.np_math_ops import rad2deg # line: 838
from tensorflow.python.ops.numpy_ops.np_math_ops import reciprocal # line: 740
from tensorflow.python.ops.numpy_ops.np_math_ops import remainder # line: 176
from tensorflow.python.ops.numpy_ops.np_math_ops import signbit # line: 746
from tensorflow.python.ops.numpy_ops.np_math_ops import sin # line: 757
from tensorflow.python.ops.numpy_ops.np_math_ops import sinc # line: 1036
from tensorflow.python.ops.numpy_ops.np_math_ops import sinh # line: 775
from tensorflow.python.ops.numpy_ops.np_math_ops import sort # line: 1385
from tensorflow.python.ops.numpy_ops.np_math_ops import sqrt # line: 692
from tensorflow.python.ops.numpy_ops.np_math_ops import square # line: 1048
from tensorflow.python.ops.numpy_ops.np_math_ops import subtract # line: 102
from tensorflow.python.ops.numpy_ops.np_math_ops import tan # line: 769
from tensorflow.python.ops.numpy_ops.np_math_ops import tanh # line: 787
from tensorflow.python.ops.numpy_ops.np_math_ops import tensordot # line: 267
from tensorflow.python.ops.numpy_ops.np_math_ops import tile # line: 1323
from tensorflow.python.ops.numpy_ops.np_math_ops import trace # line: 1504
from tensorflow.python.ops.numpy_ops.np_math_ops import true_divide # line: 120
from tensorflow.python.ops.numpy_ops.np_math_ops import vdot # line: 379
from tensorflow.python.ops.numpy_ops.np_utils import finfo # line: 486
from tensorflow.python.ops.numpy_ops.np_utils import promote_types # line: 564
from tensorflow.python.ops.numpy_ops.np_utils import result_type # line: 518



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Switching v2 features on and off."""

from tensorflow.python import tf2
from tensorflow.python.eager import monitoring
from tensorflow.python.framework import ops
from tensorflow.python.framework import registry
from tensorflow.python.framework import tensor
from tensorflow.python.framework import tensor_shape
from tensorflow.python.ops import control_flow_v2_toggles
from tensorflow.python.ops import resource_variables_toggle
from tensorflow.python.util.tf_export import tf_export

# Metrics to track the status of v2_behavior
_v2_behavior_usage_gauge = monitoring.BoolGauge(
    "/tensorflow/version/v2_behavior",
    "whether v2_behavior is enabled or disabled", "status")

_DATA_V2_CALLBACKS = registry.Registry("data_v2_callbacks")


def register_data_v2_callback(data_v2_func):
  _DATA_V2_CALLBACKS.register(data_v2_func, data_v2_func.__module__)


@tf_export(v1=["enable_v2_behavior"])
def enable_v2_behavior():
  """Enables TensorFlow 2.x behaviors.

  This function can be called at the beginning of the program (before `Tensors`,
  `Graphs` or other structures have been created, and before devices have been
  initialized. It switches all global behaviors that are different between
  TensorFlow 1.x and 2.x to behave as intended for 2.x.

  This function is called in the main TensorFlow `__init__.py` file, user should
  not need to call it, except during complex migrations.

  @compatibility(TF2)
  This function is not necessary if you are using TF2. V2 behavior is enabled by
  default.
  @end_compatibility
  """
  _v2_behavior_usage_gauge.get_cell("enable").set(True)
  # TF2 behavior is enabled if either 1) enable_v2_behavior() is called or
  # 2) the TF2_BEHAVIOR=1 environment variable is set.  In the latter case,
  # the modules below independently check if tf2.enabled().
  tf2.enable()
  ops.enable_eager_execution()
  tensor_shape.enable_v2_tensorshape()  # Also switched by tf2
  resource_variables_toggle.enable_resource_variables()
  tensor.enable_tensor_equality()
  # Enables TensorArrayV2 and control flow V2.
  control_flow_v2_toggles.enable_control_flow_v2()
  # Make sure internal uses of tf.data symbols map to V2 versions.
  for v2_enabler_name in _DATA_V2_CALLBACKS.list():
    v2_enabler = _DATA_V2_CALLBACKS.lookup(v2_enabler_name)
    v2_enabler()


@tf_export(v1=["disable_v2_behavior"])
def disable_v2_behavior():
  """Disables TensorFlow 2.x behaviors.

  This function can be called at the beginning of the program (before `Tensors`,
  `Graphs` or other structures have been created, and before devices have been
  initialized. It switches all global behaviors that are different between
  TensorFlow 1.x and 2.x to behave as intended for 1.x.

  User can call this function to disable 2.x behavior during complex migrations.

  @compatibility(TF2)
  Using this function indicates that your software is not compatible
  with eager execution and `tf.function` in TF2.

  To migrate to TF2, rewrite your code to be compatible with eager execution.
  Please refer to the [migration guide]
  (https://www.tensorflow.org/guide/migrate) for additional resource on the
  topic.
  @end_compatibility
  """
  _v2_behavior_usage_gauge.get_cell("disable").set(True)
  tf2.disable()
  ops.disable_eager_execution()
  tensor_shape.disable_v2_tensorshape()  # Also switched by tf2
  resource_variables_toggle.disable_resource_variables()
  tensor.disable_tensor_equality()
  # Disables TensorArrayV2 and control flow V2.
  control_flow_v2_toggles.disable_control_flow_v2()
  # Make sure internal uses of tf.data symbols map to V1 versions.
  for v2_disabler_name in _DATA_V2_CALLBACKS.list():
    v2_disabler = _DATA_V2_CALLBACKS.lookup(v2_disabler_name)
    v2_disabler()



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for the trace_type module."""

from typing import Any, List, Tuple

import numpy as np


# TODO(b/225045380): Depend on the abstracted `leaf` lib from 'nest'.
def is_namedtuple(obj):
  return hasattr(obj, "_fields") and all(
      isinstance(field, str) for field in obj._fields)


# TODO(b/225045380): Depend on the abstracted `leaf` lib from 'nest'.
def is_attrs(obj):
  return hasattr(type(obj), "__attrs_attrs__")


# TODO(b/225045380): Depend on the abstracted `leaf` lib from 'nest'.
def is_np_ndarray(value):
  return hasattr(value, "__array__") and not (
      # For legacy reasons we do not automatically promote Numpy strings.
      isinstance(value, np.str_)
      # NumPy dtypes have __array__ as unbound methods.
      or isinstance(value, type))


def cast_and_return_whether_casted(
    trace_types, values, context
) -> Tuple[List[Any], bool]:
  did_cast = False
  casted_values = []
  for t, v in zip(trace_types, values):
    casted_v = t.cast(v, context)
    casted_values.append(casted_v)
    if casted_v is not v:
      did_cast = True
  return casted_values, did_cast



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.__internal__.tf2 namespace
"""

import sys as _sys

from tensorflow.python.tf2 import enabled # line: 35



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.experimental namespace
"""

import sys as _sys

from tensorflow._api.v2.compat.v2.experimental import dlpack
from tensorflow._api.v2.compat.v2.experimental import dtensor
from tensorflow._api.v2.compat.v2.experimental import extension_type
from tensorflow._api.v2.compat.v2.experimental import numpy
from tensorflow._api.v2.compat.v2.experimental import tensorrt
from tensorflow.python.data.ops.optional_ops import Optional # line: 31
from tensorflow.python.eager.context import async_clear_error # line: 3118
from tensorflow.python.eager.context import async_scope # line: 3054
from tensorflow.python.eager.context import function_executor_type # line: 2903
from tensorflow.python.framework.dtypes import float8_e4m3b11fnuz # line: 465
from tensorflow.python.framework.dtypes import float8_e4m3fn # line: 439
from tensorflow.python.framework.dtypes import float8_e4m3fnuz # line: 452
from tensorflow.python.framework.dtypes import float8_e5m2 # line: 426
from tensorflow.python.framework.dtypes import float8_e5m2fnuz # line: 478
from tensorflow.python.framework.dtypes import int2 # line: 496
from tensorflow.python.framework.dtypes import int4 # line: 484
from tensorflow.python.framework.dtypes import uint2 # line: 502
from tensorflow.python.framework.dtypes import uint4 # line: 490
from tensorflow.python.framework.extension_type import BatchableExtensionType # line: 831
from tensorflow.python.framework.extension_type import ExtensionType # line: 99
from tensorflow.python.framework.extension_type import ExtensionTypeBatchEncoder # line: 651
from tensorflow.python.framework.extension_type import ExtensionTypeSpec # line: 419
from tensorflow.python.framework.load_library import register_filesystem_plugin # line: 199
from tensorflow.python.framework.strict_mode import enable_strict_mode # line: 22
from tensorflow.python.ops.ragged.dynamic_ragged_shape import DynamicRaggedShape # line: 167
from tensorflow.python.ops.ragged.row_partition import RowPartition # line: 56
from tensorflow.python.ops.structured.structured_tensor import StructuredTensor # line: 59
from tensorflow.python.util.dispatch import dispatch_for_api # line: 334
from tensorflow.python.util.dispatch import dispatch_for_binary_elementwise_apis # line: 878
from tensorflow.python.util.dispatch import dispatch_for_binary_elementwise_assert_apis # line: 936
from tensorflow.python.util.dispatch import dispatch_for_unary_elementwise_apis # line: 811
from tensorflow.python.util.dispatch import unregister_dispatch_for # line: 553



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.__internal__.ops namespace
"""

import sys as _sys

from tensorflow.python.ops.handle_data_util import get_resource_handle_data # line: 25
from tensorflow.python.ops.resource_variable_ops import is_resource_variable # line: 2566
from tensorflow.python.ops.weights_broadcast_ops import broadcast_weights # line: 136



# Copyright 2023 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for strategies that are backed by DTensor."""

from tensorflow.dtensor.python import accelerator_util
from tensorflow.dtensor.python import api as d_api
from tensorflow.dtensor.python import input_util
from tensorflow.dtensor.python import layout
from tensorflow.python.distribute import cross_device_ops as cross_device_ops_lib
from tensorflow.python.distribute import device_util
from tensorflow.python.distribute import distribute_lib
from tensorflow.python.distribute import reduce_util
from tensorflow.python.distribute import values as values_lib
from tensorflow.python.eager import context
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_conversion_registry
from tensorflow.python.framework import tensor_util
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import summary_ops_v2


# Default dimension name used for the mesh created when user provide a list
# of devices. For mirrored strategy, it should be a 1D mesh with batch dim only.
DEFAULT_BATCH_MESH_DIM_NAME = "batch"


class DTensorDistributedValue(values_lib.DistributedValues):
  """DistributedValue backed by a DTensor instance.

  This class is useful to align the interface between DTensor and tf.distribute.
  Most of the tf.distribute API will accept/return DistributedValue, whereas
  DTensor low level API will only accept DTensor instance. In order to avoid
  the conversion back and forth between DistributedValue and DTensor, we
  introduce this class so that it can work with both side.
  """

  def __init__(self, dtensor):
    if context.executing_eagerly():
      if not d_api.is_dtensor(dtensor):
        raise ValueError("The DTensorDistributedValue can only be built with "
                         f"DTensor instance, got {type(dtensor)}")
      super().__init__(d_api.unpack(dtensor))
    else:
      # We can't unpack the dtensor instance for now due to graph context.
      # We will treat the dtensor instance as one global instance and let it
      # return as a global replica instance.
      # TODO(feyu): Support unpack in the graph context.
      super().__init__([dtensor,])
    self._dtensor = dtensor

  def get_dtensor(self):
    return self._dtensor

  @property
  def values(self):
    # Note that this method exists so that it match the interface for PerReplica
    # The public API in `tf.types.experimental.distributed.PerReplica` doesn't
    # define any methods.
    return self._values


def _dtensor_distributed_value_to_tensor(
    var, dtype=None, name=None, as_ref=False):
  del name
  dtensor = var.get_dtensor()
  if dtype is not None and not dtype.is_compatible_with(dtensor.dtype):
    raise ValueError(
        "Incompatible type conversion requested to type {!r} for variable "
        "of type {!r}".format(dtype.name, dtensor.dtype.name))
  if as_ref:
    raise NotImplementedError(
        "PerReplica doesn't support being used as a reference.")
  return dtensor


# Register a conversion function to provide a useful error message when users
# try to use PerReplica values in the wrong contexts
tensor_conversion_registry.register_tensor_conversion_function(
    DTensorDistributedValue, _dtensor_distributed_value_to_tensor)


class DTensorReplicaContext(distribute_lib.ReplicaContext):
  """ReplicaContext for strategy that is backed by DTensor.

  Since the DTensor is operated in the global context, most of the methods from
  existing strategy ReplicaContext is not applicable since they need to access
  local values. For now most of the methods in this class will raise explicit
  error to user, and we will add more support for local values in future.
  """
  _UNSUPPORTED_ERROR_MSG = (
      "Strategy that is backed by DTensor is run with a global context, and "
      "doesn't support operations for local context, like any call to merge/"
      "gather/reduce or local replica ID. Please use any strategy that is not "
      "backed by DTensor")

  def __init__(self, strategy):
    # Since DTensor strategy only runs in a global context, and we can't have
    # a local replica ID in the sync group. For now we pass None to parent, and
    # raise an explicit error when it is accessed.
    super().__init__(strategy, replica_id_in_sync_group=None)

  def __enter__(self):
    # This is a copy of parent class, without any check about whether the
    # current replica is the first one (since DTensor only has one).
    distribute_lib._push_per_thread_mode(self._thread_context)  # # pylint: disable=protected-access

    summary_state = summary_ops_v2._summary_state  # pylint: disable=protected-access
    self._summary_recording_distribution_strategy = (
        summary_state.is_recording_distribution_strategy)
    summary_state.is_recording_distribution_strategy = True

  @property
  def replica_id_in_sync_group(self):
    # Since there is only one global context for DTensor, we always return a
    # constant value here. This value is needed by the RNG which try to generate
    # different seed for different replica.
    return 0

  @property
  def _replica_id(self):
    raise NotImplementedError(self._UNSUPPORTED_ERROR_MSG)

  def merge_call(self, merge_fn, args=(), kwargs=None):
    raise NotImplementedError(self._UNSUPPORTED_ERROR_MSG)

  def all_reduce(self, reduce_op, value, options=None):
    raise NotImplementedError(self._UNSUPPORTED_ERROR_MSG)

  def all_gather(self, value, axis, options=None):
    raise NotImplementedError(self._UNSUPPORTED_ERROR_MSG)

  def _update(self, var, fn, args=(), kwargs=None, group=True):
    raise NotImplementedError(self._UNSUPPORTED_ERROR_MSG)


def initialize_accelerator_system_once(device_type):
  # Initialize the GPU/TPU before creating the mesh.
  # Note that this method will also trigger the creation of the pairing
  # virtual host CPUs, which is needed by dataset and checkpoint.
  if not accelerator_util.is_initialized():
    # TODO(feyu): Add a method in accelerator_util to check the initialized
    # mesh device types.
    accelerator_util.initialize_accelerator_system(
        device_type,
        experimental_reset_context=True)


def convert_inputs_to_dtensor(inputs, mesh):
  """Convert any input types to DTensor instance."""
  if isinstance(inputs, DTensorDistributedValue):
    return inputs.get_dtensor()
  elif isinstance(inputs, values_lib.DistributedValues):
    return convert_per_replica_to_dtensor(inputs, mesh)
  elif isinstance(inputs, input_util._DTensorIterator):   # pylint: disable=protected-access
    return inputs
  elif tensor_util.is_tensor(inputs):
    if context.executing_eagerly():
      if d_api.is_dtensor(inputs):
        return inputs
      else:
        # For a non-dtensor input in eager context, we could choose to replica
        # them into per-replica and then pack them into dtensor. However, this
        # will cause an eager/graph discrepancy since we can't do this check in
        # the graph context. For now, we will ask user to provide a distributed
        # value for inputs.
        _raise_unsupported_input_type_error(inputs)
    else:
      # For graph context, since we can't check if they are dtensor or not. We
      # will assume the value is already distributed. This is a critical use
      # case for keras, where all the inputs are pre-distributed via strategy,
      # and the train function execute within graph context.
      return inputs
  else:
    # For any other types.
    _raise_unsupported_input_type_error(inputs)


def _raise_unsupported_input_type_error(inputs):
  raise ValueError("Unsupported input types for MirroredStrategy. "
                   "Please use `strategy.distribute_dataset` or "
                   "`strategy.distribute_values_from_function` to "
                   f"distribute inputs. Received input type: {type(inputs)}")


def is_distributed_value(value):
  return isinstance(
      value, values_lib.DistributedValues) or d_api.is_dtensor(value)


def convert_per_replica_to_dtensor(per_replica_value, mesh):
  """Convert a PerReplica result to a DTensor instance.

  Args:
    per_replica_value: A PerReplica instance whose value will be converted
      to DTensor.
    mesh: The mesh used for layout creation.

  Returns:
    A DTensor instance that packed from per_replica_value with batch sharded
      layout.
  """
  values = per_replica_value.values
  if isinstance(values[0], (float, int)):
    rank = 0
  else:
    rank = len(values[0].shape)

  if rank == 0:
    result = []
    # dtensor.pack requires each component to have same rank as the packed
    # result. When the individual value is scalar, it needs to be expanded into
    # 1D tensor.
    for v in values:
      result.append(array_ops.expand_dims_v2(v, axis=0))
    rank += 1
  else:
    result = list(values)   # dtensor.pack requires a list as input.

  # TODO(scottzhu): Note that the result tensor could be a partial value and
  # not always batch shard or fully replicaed. See
  # http://screenshot/6ERkXyX95KqftCw as an example.
  batch_layout = layout.Layout.batch_sharded(
      mesh, batch_dim=DEFAULT_BATCH_MESH_DIM_NAME, rank=rank)

  return d_api.pack(result, batch_layout)


def dtensor_reduce(strategy, reduce_op, value, axis):
  """Implement dtensor based strategy.reduce()."""
  # Due to the limitation of using scalar in DTensor (e.g. the rank 0 tensor
  # loss the batch shard information), we need to override the default
  # reduce in addition to the strategy.extend._reduce_to()
  # Most of the logic here is a mimic of the parent class, except for how
  # mean and sum are calculated in a global context.
  distribute_lib._require_cross_replica_or_default_context_extended(  # pylint: disable=protected-access
      strategy.extended)
  if isinstance(reduce_op, str):
    reduce_op = reduce_util.ReduceOp(reduce_op.upper())

  distributed_input = is_distributed_value(value)
  if not distributed_input and axis is None:
    # For any value that isn't distributed and doesn't need a reduction within
    # the replica.
    destinations = (device_util.current() or
                    strategy.extended._default_device or  # pylint: disable=protected-access
                    "/device:CPU:0")
    devices = cross_device_ops_lib.get_devices_from(destinations)
    with ops.device(devices[0]):
      return array_ops.identity(
          cross_device_ops_lib.reduce_non_distributed_value(
              reduce_op, value, destinations, strategy.num_replicas_in_sync))

  value = convert_inputs_to_dtensor(value, strategy._mesh)  # pylint: disable=protected-access
  # At this point, the value is a DTensor instance now.
  # There will be a final reduction step cross replica. In order to maintain
  # the shape of each local replica, we need to add a new dim to the front.
  # E.g. 2 replica with local shape as (4, 5, 6), the global tensor shape
  # should be (8, 5, 6), we will reshape into (2, 4, 5, 6) and then do a
  # reduction on axis 0.
  if reduce_op == reduce_util.ReduceOp.MEAN:
    reduce_op = math_ops.reduce_mean
  else:
    reduce_op = math_ops.reduce_sum

  # TODO(scottzhu): Make sure we handle dynamic/uneven shape in future.
  if d_api.fetch_layout(value).is_fully_replicated():
    # In case of fully mirrored dtensor, we only need to do one reduce, and
    # don't need to care about any per-replica logic.
    if axis is not None:
      value = reduce_op(value, axis=axis)
  else:
    new_shape = [strategy.num_replicas_in_sync, -1]
    if len(value.shape) > 1:
      new_shape.extend(array_ops.shape(value)[1:])
    value = array_ops.reshape(value, new_shape)
    if axis is not None:
      # we do a reduce_sum/mean within each of the replica when axis is not
      # None. Add 1 to the axis since there is a new dim added by reshape in
      # front.
      value = reduce_op(value, axis=axis + 1)
    value = reduce_op(value, axis=0)

  # Note that we return a DTensor instance here, which should have the same
  # value as the original MirroredStrategy, but with a different type. User
  # might want a tf.Tensor for the status quo.
  return value



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.distribute namespace
"""

import sys as _sys

from tensorflow._api.v2.distribute import cluster_resolver
from tensorflow._api.v2.distribute import coordinator
from tensorflow._api.v2.distribute import experimental
from tensorflow.python.distribute.collective_all_reduce_strategy import CollectiveAllReduceStrategy as MultiWorkerMirroredStrategy # line: 56
from tensorflow.python.distribute.cross_device_ops import CrossDeviceOps # line: 251
from tensorflow.python.distribute.cross_device_ops import HierarchicalCopyAllReduce # line: 996
from tensorflow.python.distribute.cross_device_ops import NcclAllReduce # line: 959
from tensorflow.python.distribute.cross_device_ops import ReductionToOneDevice # line: 581
from tensorflow.python.distribute.distribute_lib import InputContext # line: 840
from tensorflow.python.distribute.distribute_lib import InputOptions # line: 1014
from tensorflow.python.distribute.distribute_lib import InputReplicationMode # line: 823
from tensorflow.python.distribute.distribute_lib import ReplicaContext # line: 3669
from tensorflow.python.distribute.distribute_lib import RunOptions # line: 979
from tensorflow.python.distribute.distribute_lib import Strategy # line: 2025
from tensorflow.python.distribute.distribute_lib import StrategyExtendedV2 as StrategyExtended # line: 2393
from tensorflow.python.distribute.distribute_lib import experimental_set_strategy # line: 583
from tensorflow.python.distribute.distribute_lib import get_replica_context # line: 453
from tensorflow.python.distribute.distribute_lib import get_strategy # line: 542
from tensorflow.python.distribute.distribute_lib import has_strategy # line: 562
from tensorflow.python.distribute.distribute_lib import in_cross_replica_context # line: 517
from tensorflow.python.distribute.mirrored_strategy import MirroredStrategy # line: 199
from tensorflow.python.distribute.one_device_strategy import OneDeviceStrategy # line: 38
from tensorflow.python.distribute.parameter_server_strategy_v2 import ParameterServerStrategyV2 as ParameterServerStrategy # line: 72
from tensorflow.python.distribute.reduce_util import ReduceOp # line: 23
from tensorflow.python.distribute.tpu_strategy import TPUStrategyV2 as TPUStrategy # line: 242
from tensorflow.python.training.server_lib import Server # line: 94
from tensorflow.python.types.distribute import DistributedDatasetInterface as DistributedDataset # line: 307
from tensorflow.python.types.distribute import DistributedIteratorInterface as DistributedIterator # line: 196
from tensorflow.python.types.distribute import DistributedValues # line: 62



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.autograph.experimental namespace
"""

import sys as _sys

from tensorflow.python.autograph.core.converter import Feature # line: 78
from tensorflow.python.autograph.impl.api import do_not_convert # line: 624
from tensorflow.python.autograph.lang.directives import set_loop_options # line: 45



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tensorflow/core/framework/dataset_options.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'tensorflow/core/framework/dataset_options.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorflow.core.framework import model_pb2 as tensorflow_dot_core_dot_framework_dot_model__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/tensorflow/core/framework/dataset_options.proto\x12\x0ftensorflow.data\x1a%tensorflow/core/framework/model.proto\"\xef\x02\n\x0f\x41utotuneOptions\x12\x11\n\x07\x65nabled\x18\x01 \x01(\x08H\x00\x12\x14\n\ncpu_budget\x18\x02 \x01(\x05H\x01\x12\x14\n\nram_budget\x18\x03 \x01(\x03H\x02\x12\x46\n\x12\x61utotune_algorithm\x18\x04 \x01(\x0e\x32(.tensorflow.data.model.AutotuneAlgorithmH\x03\x12\x1d\n\x13initial_parallelism\x18\x05 \x01(\x03H\x04\x12\x19\n\x0fmin_parallelism\x18\x06 \x01(\x03H\x05\x42\x12\n\x10optional_enabledB\x15\n\x13optional_cpu_budgetB\x15\n\x13optional_ram_budgetB\x1d\n\x1boptional_autotune_algorithmB\x1e\n\x1coptional_initial_parallelismB\x1a\n\x18optional_min_parallelism\"\xd1\x01\n\x12\x43\x61rdinalityOptions\x12G\n\rcompute_level\x18\x01 \x01(\x0e\x32\x30.tensorflow.data.CardinalityOptions.ComputeLevel\"r\n\x0c\x43omputeLevel\x12#\n\x1f\x43\x41RDINALITY_COMPUTE_UNSPECIFIED\x10\x00\x12\x1b\n\x17\x43\x41RDINALITY_COMPUTE_LOW\x10\x01\x12 \n\x1c\x43\x41RDINALITY_COMPUTE_MODERATE\x10\x02\"\x7f\n\x11\x44istributeOptions\x12;\n\x11\x61uto_shard_policy\x18\x01 \x01(\x0e\x32 .tensorflow.data.AutoShardPolicy\x12\x15\n\x0bnum_devices\x18\x02 \x01(\x05H\x00\x42\x16\n\x14optional_num_devices\"\xb9\x06\n\x13OptimizationOptions\x12%\n\x1b\x61pply_default_optimizations\x18\x01 \x01(\x08H\x00\x12\x17\n\rfilter_fusion\x18\x06 \x01(\x08H\x01\x12\x1e\n\x14map_and_batch_fusion\x18\t \x01(\x08H\x02\x12\x1f\n\x15map_and_filter_fusion\x18\n \x01(\x08H\x03\x12\x14\n\nmap_fusion\x18\x0b \x01(\x08H\x04\x12\x1d\n\x13map_parallelization\x18\x0c \x01(\x08H\x05\x12\x1a\n\x10noop_elimination\x18\x0e \x01(\x08H\x06\x12\x18\n\x0eparallel_batch\x18\x0f \x01(\x08H\x07\x12#\n\x19shuffle_and_repeat_fusion\x18\x11 \x01(\x08H\x08\x12 \n\x16\x66ilter_parallelization\x18\x12 \x01(\x08H\t\x12\x19\n\x0finject_prefetch\x18\x13 \x01(\x08H\n\x12!\n\x17seq_interleave_prefetch\x18\x15 \x01(\x08H\x0b\x42&\n$optional_apply_default_optimizationsB\x18\n\x16optional_filter_fusionB\x1f\n\x1doptional_map_and_batch_fusionB \n\x1eoptional_map_and_filter_fusionB\x15\n\x13optional_map_fusionB\x1e\n\x1coptional_map_parallelizationB\x1b\n\x19optional_noop_eliminationB\x19\n\x17optional_parallel_batchB$\n\"optional_shuffle_and_repeat_fusionB!\n\x1foptional_filter_parallelizationB\x1a\n\x18optional_inject_prefetchB\"\n optional_seq_interleave_prefetchJ\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\tJ\x04\x08\r\x10\x0eJ\x04\x08\x10\x10\x11J\x04\x08\x14\x10\x15\"5\n\x0eServiceOptions\x12\x10\n\x06pinned\x18\x01 \x01(\x08H\x00\x42\x11\n\x0foptional_pinned\"\xa2\x01\n\x10ThreadingOptions\x12\"\n\x18max_intra_op_parallelism\x18\x01 \x01(\x05H\x00\x12!\n\x17private_threadpool_size\x18\x02 \x01(\x05H\x01\x42#\n!optional_max_intra_op_parallelismB\"\n optional_private_threadpool_size\"\xb5\x05\n\x07Options\x12\x16\n\x0c\x64\x61taset_name\x18\n \x01(\tH\x00\x12\x16\n\x0e\x66ramework_type\x18\x0b \x03(\t\x12\x17\n\rdeterministic\x18\x01 \x01(\x08H\x01\x12:\n\x10\x61utotune_options\x18\x07 \x01(\x0b\x32 .tensorflow.data.AutotuneOptions\x12>\n\x12\x64istribute_options\x18\x02 \x01(\x0b\x32\".tensorflow.data.DistributeOptions\x12\x42\n\x14optimization_options\x18\x03 \x01(\x0b\x32$.tensorflow.data.OptimizationOptions\x12\x38\n\x0fservice_options\x18\x0c \x01(\x0b\x32\x1f.tensorflow.data.ServiceOptions\x12\x0f\n\x05slack\x18\x04 \x01(\x08H\x02\x12<\n\x11threading_options\x18\x05 \x01(\x0b\x32!.tensorflow.data.ThreadingOptions\x12\x45\n\x15\x65xternal_state_policy\x18\x06 \x01(\x0e\x32$.tensorflow.data.ExternalStatePolicyH\x03\x12\x1d\n\x13symbolic_checkpoint\x18\x08 \x01(\x08H\x04\x12\x14\n\nwarm_start\x18\t \x01(\x08H\x05\x42\x17\n\x15optional_dataset_nameB\x18\n\x16optional_deterministicB\x10\n\x0eoptional_slackB \n\x1eoptional_external_state_policyB\x1e\n\x1coptional_symbolic_checkpointB\x15\n\x13optional_warm_start*K\n\x0f\x41utoShardPolicy\x12\x08\n\x04\x41UTO\x10\x00\x12\x08\n\x04\x46ILE\x10\x01\x12\x08\n\x04\x44\x41TA\x10\x02\x12\x08\n\x04HINT\x10\x03\x12\x10\n\x03OFF\x10\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01*J\n\x13\x45xternalStatePolicy\x12\x0f\n\x0bPOLICY_WARN\x10\x00\x12\x11\n\rPOLICY_IGNORE\x10\x01\x12\x0f\n\x0bPOLICY_FAIL\x10\x02\x42XZVgithub.com/tensorflow/tensorflow/tensorflow/go/core/framework/dataset_options_go_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tensorflow.core.framework.dataset_options_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'ZVgithub.com/tensorflow/tensorflow/tensorflow/go/core/framework/dataset_options_go_proto'
  _globals['_AUTOSHARDPOLICY']._serialized_start=2562
  _globals['_AUTOSHARDPOLICY']._serialized_end=2637
  _globals['_EXTERNALSTATEPOLICY']._serialized_start=2639
  _globals['_EXTERNALSTATEPOLICY']._serialized_end=2713
  _globals['_AUTOTUNEOPTIONS']._serialized_start=108
  _globals['_AUTOTUNEOPTIONS']._serialized_end=475
  _globals['_CARDINALITYOPTIONS']._serialized_start=478
  _globals['_CARDINALITYOPTIONS']._serialized_end=687
  _globals['_CARDINALITYOPTIONS_COMPUTELEVEL']._serialized_start=573
  _globals['_CARDINALITYOPTIONS_COMPUTELEVEL']._serialized_end=687
  _globals['_DISTRIBUTEOPTIONS']._serialized_start=689
  _globals['_DISTRIBUTEOPTIONS']._serialized_end=816
  _globals['_OPTIMIZATIONOPTIONS']._serialized_start=819
  _globals['_OPTIMIZATIONOPTIONS']._serialized_end=1644
  _globals['_SERVICEOPTIONS']._serialized_start=1646
  _globals['_SERVICEOPTIONS']._serialized_end=1699
  _globals['_THREADINGOPTIONS']._serialized_start=1702
  _globals['_THREADINGOPTIONS']._serialized_end=1864
  _globals['_OPTIONS']._serialized_start=1867
  _globals['_OPTIONS']._serialized_end=2560
# @@protoc_insertion_point(module_scope)



# Copyright 2022 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for extracting and writing checkpoint info`."""

from tensorflow.core.protobuf import trackable_object_graph_pb2
from tensorflow.python.ops import resource_variable_ops
from tensorflow.python.ops import variables
from tensorflow.python.trackable import trackable_utils
from tensorflow.python.util import object_identity


def serialize_slot_variables(trackable_objects, node_ids, object_names):
  """Gather and name slot variables."""
  non_slot_objects = list(trackable_objects)
  slot_variables = object_identity.ObjectIdentityDictionary()
  for trackable in non_slot_objects:
        # TODO(b/110718070): Fix Keras imports.
        # Note: dir() is used rather than hasattr() here to avoid triggering
        # custom __getattr__ code, see b/152031870 for context.
    if "get_slot_names" in dir(trackable):
      slot_names = trackable.get_slot_names()
      for slot_name in slot_names:
        for original_variable_node_id, original_variable in enumerate(
            non_slot_objects):
          try:
            slot_variable = trackable.get_slot(original_variable, slot_name)
          except (AttributeError, KeyError):
            slot_variable = None
          if slot_variable is None:
            continue
          slot_variable._maybe_initialize_trackable()  # pylint: disable=protected-access
          if slot_variable._trackable_children():  # pylint: disable=protected-access
            # TODO(allenl): Gather dependencies of slot variables.
            raise NotImplementedError(
                "Currently only variables with no dependencies can be saved as "
                "slot variables. File a feature request if this limitation "
                "bothers you.")
          if slot_variable in node_ids:
            raise NotImplementedError(
                "A slot variable was re-used as a dependency of a Trackable "
                f"object: {slot_variable}. This is not currently allowed. "
                "File a feature request if this limitation bothers you.")
          checkpoint_name = trackable_utils.slot_variable_key(
              variable_path=object_names[original_variable],
              optimizer_path=object_names[trackable],
              slot_name=slot_name)
          object_names[slot_variable] = checkpoint_name
          slot_variable_node_id = len(trackable_objects)
          node_ids[slot_variable] = slot_variable_node_id
          trackable_objects.append(slot_variable)
          slot_variable_proto = (
              trackable_object_graph_pb2.TrackableObjectGraph.TrackableObject
              .SlotVariableReference(
                  slot_name=slot_name,
                  original_variable_node_id=original_variable_node_id,
                  slot_variable_node_id=slot_variable_node_id))
          slot_variables.setdefault(trackable, []).append(slot_variable_proto)
  return slot_variables


def get_mapped_trackable(trackable, object_map):
  """Returns the mapped trackable if possible, otherwise returns trackable."""
  if object_map is None:
    return trackable
  else:
    return object_map.get(trackable, trackable)


def get_full_name(var):
  """Gets the full name of variable for name-based checkpoint compatibility."""
  # pylint: disable=protected-access
  if (not (isinstance(var, variables.Variable) or
           # Some objects do not subclass Variable but still act as one.
           resource_variable_ops.is_resource_variable(var))):
    return ""

  if getattr(var, "_save_slice_info", None) is not None:
    # Use getattr because `var._save_slice_info` may be set as `None`.
    return var._save_slice_info.full_name
  else:
    return var._shared_name
  # pylint: enable=protected-access


def add_checkpoint_values_check(object_graph_proto):
  """Determines which objects have checkpoint values and save this to the proto.

  Args:
    object_graph_proto: A `TrackableObjectGraph` proto.
  """
  # Trackable -> set of all trackables that depend on it (the "parents").
  # If a trackable has checkpoint values, then all of the parents can be
  # marked as having checkpoint values.
  parents = {}
  checkpointed_trackables = object_identity.ObjectIdentitySet()

  # First pass: build dictionary of parent objects and initial set of
  # checkpointed trackables.
  checkpointed_trackables = set()
  for node_id, object_proto in enumerate(object_graph_proto.nodes):
    if (object_proto.attributes or object_proto.slot_variables or
        object_proto.HasField("registered_saver")):
      checkpointed_trackables.add(node_id)
    for child_proto in object_proto.children:
      child = child_proto.node_id
      if child not in parents:
        parents[child] = set()
      parents[child].add(node_id)

  # Second pass: add all connected parents to set of checkpointed trackables.
  to_visit = set()
  to_visit.update(checkpointed_trackables)

  while to_visit:
    trackable = to_visit.pop()
    if trackable not in parents:
      # Some trackables may not have parents (e.g. slot variables).
      continue
    current_parents = parents.pop(trackable)
    checkpointed_trackables.update(current_parents)
    for parent in current_parents:
      if parent in parents:
        to_visit.add(parent)

  for node_id, object_proto in enumerate(object_graph_proto.nodes):
    object_proto.has_checkpoint_values.value = bool(
        node_id in checkpointed_trackables)


def objects_ids_and_slot_variables_and_paths(graph_view,
                                             skip_slot_variables=False):
  """Traverse the object graph and list all accessible objects.

  Looks for `Trackable` objects which are dependencies of
  `root_trackable`. Includes slot variables only if the variable they are
  slotting for and the optimizer are dependencies of `root_trackable`
  (i.e. if they would be saved with a checkpoint).

  Args:
    graph_view: A GraphView object.
    skip_slot_variables: If True does not return trackables for slot variable.
      Default False.

  Returns:
    A tuple of (trackable objects, paths from root for each object,
                object -> node id, slot variables, object_names)
  """
  trackable_objects, node_paths = graph_view.breadth_first_traversal()
  object_names = object_identity.ObjectIdentityDictionary()
  for obj, path in node_paths.items():
    object_names[obj] = trackable_utils.object_path_to_string(path)
  node_ids = object_identity.ObjectIdentityDictionary()
  for node_id, node in enumerate(trackable_objects):
    node_ids[node] = node_id
  if skip_slot_variables:
    slot_variables = object_identity.ObjectIdentityDictionary()
  else:
    slot_variables = serialize_slot_variables(
        trackable_objects=trackable_objects,
        node_ids=node_ids,
        object_names=object_names,
    )
  return (trackable_objects, node_paths, node_ids, slot_variables, object_names)


def list_objects(graph_view, skip_slot_variables=False):
  """Traverse the object graph and list all accessible objects."""
  trackable_objects = objects_ids_and_slot_variables_and_paths(
      graph_view, skip_slot_variables
  )[0]
  return trackable_objects



# Copyright 2023 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implement a StrategyExtended based on the DTensor low level API."""

import functools

from tensorflow.dtensor.python import api as d_api
from tensorflow.dtensor.python import config as d_config
from tensorflow.dtensor.python import d_variable
from tensorflow.dtensor.python import input_util
from tensorflow.dtensor.python import layout
from tensorflow.python.data.experimental.ops import distribute
from tensorflow.python.distribute import distribute_lib
from tensorflow.python.distribute import distribute_utils
from tensorflow.python.distribute.experimental import dtensor_util
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.util import nest


class DTensorStrategyExtended(distribute_lib.StrategyExtendedV2):
  """Strategy extension that support both single and multi worker strategy."""
  # Note that the unit test for this class is via the strategy interface.

  def __init__(self, container_strategy, mesh):
    super().__init__(container_strategy)
    self._mesh = mesh
    self._num_clients = d_config.num_clients()
    self._client_id = d_config.client_id()

  def _create_variable(self, next_creator, **kwargs):
    # Make sure the pop the `use_resource` which is not supported by the
    # base tf.Variable. The `use_resource` is added by
    # creator_with_resource_vars in distribute_lib.py
    kwargs.pop('use_resource', None)

    # Ignore the colocate_with for the mirrored strategy. Each of the device
    # will get same copy of variable in the DTensor's case.
    # `colocate_with` is added when user call:
    # strategy.extended.colocate_vars_with(variable)
    kwargs.pop('colocate_with', None)

    # Ignore expected_shape, which is from the v1 Variable. Keras was somehow
    # using the v1 Variable, but didn't specify that value particularly.
    kwargs.pop('expected_shape', None)

    # Make sure to call DVariable initializer under the scope so that it will
    # have the proper replicated layout. The initial_value is multi-typed,
    # eg it can be a tensor, or a python/numpy type, or a callable that
    # produce tensor/python/numpy types. In all those cases, we need to wrap
    # them invoke convert_to_tensor() under the scope so that the proper
    # layout can be assigned.

    # TODO(scottzhu): The layout information should be injected via kwargs, or
    # lazily set later.
    initial_value = kwargs.pop('initial_value')
    dtype = kwargs.get('dtype', None)
    def new_initial_value():
      if callable(initial_value):
        init_var = ops.convert_to_tensor(initial_value(), dtype=dtype)
      else:
        init_var = ops.convert_to_tensor(initial_value, dtype=dtype)
      rank = init_var.shape.rank
      return d_api.copy_to_mesh(
          init_var, layout.Layout.replicated(self._mesh, rank))

    return d_variable.DVariable(new_initial_value, **kwargs)

  @property
  def _num_replicas_in_sync(self):
    # The mesh should be 1D with batch sharding only.
    # In the model parallel case, it should only return the size of
    # batch dimension.
    return self._mesh.size

  def value_container(self, value):
    return value

  @property
  def worker_devices(self):
    # Note that in either single worker (MirroredStrategy) or multi worker (
    # MultiWorkerMirroredStrategy), worker_devices refers to the local worker
    # devices.
    return tuple(self._mesh.local_devices())

  @property
  def parameter_devices(self):
    # Same as the worker_devices.
    return self.worker_devices

  def _in_multi_worker_mode(self):
    return d_config.num_clients() > 1

  def _get_local_replica_id(self, replica_id_in_sync_group):
    return replica_id_in_sync_group

  def _default_device_scope(self):
    return d_api.default_mesh(self._mesh)

  def _experimental_distribute_dataset(self, dataset, options):
    # Strategy always assume the user input data is a batched dataset for
    # experimental_distribute_dataset().
    # TODO(yuefengz): Add check for whether a dataset is batched for all
    # strategies.

    # TODO(b/265198795): Support dataset already batched to global batch size.
    # Since DTensorDataset doesn't support batched dataset that is already
    # batched global batch size, it only supports dataset that is batched to
    # local batch size, we need to infer the batch size, and unbatch the dataset
    # until the b/265198795 is resolved.
    batch_size = distribute.compute_batch_size(dataset)

    # There are multiple case that the batch is not static, eg partial batch,
    # or uneven batch, in all those case, it will return -1.
    if batch_size.numpy() < 0:
      # When we don't have a static batch size.
      raise ValueError('DTensor strategy requires a static batch size for now.'
                       'The dynamic batch size will be supported in future')
    # Unbatch the dataset for now since the DTensorDataset has some limitation
    # about the local batch size as well as the mesh size.
    dataset = dataset.unbatch()

    def _create_batch_layout(tensor_spec):
      # For unbatched dataset, the new layout need to have +1 rank for
      # the batched result.
      rank = len(tensor_spec.shape) + 1
      return layout.Layout.batch_sharded(
          self._mesh, batch_dim=dtensor_util.DEFAULT_BATCH_MESH_DIM_NAME,
          rank=rank)

    layouts = nest.map_structure(_create_batch_layout, dataset.element_spec)

    return input_util.DTensorDataset(
        dataset=dataset,
        mesh=self._mesh,
        layouts=layouts,
        global_batch_size=batch_size,
        dataset_already_batched=False,
        batch_dim=dtensor_util.DEFAULT_BATCH_MESH_DIM_NAME,
        # TODO(scottzhu): Add prefetch support by inspecting the input dataset.
        prefetch=None,
        tf_data_service_config=None
    )

  def _make_dataset_iterator(self, dataset):
    raise NotImplementedError(
        'Strategy.make_dataset_iterator() is deprecated, and only available '
        'in the V1 API.')

  def _make_input_fn_iterator(self, input_fn, replication_mode):
    raise NotImplementedError(
        'Strategy.make_input_fn_iterator() is deprecated, and only available '
        'in the V1 API.')

  def _distribute_datasets_from_function(self, dataset_fn, options):
    # TODO(scottzhu): Implement the logic for options in future
    del options
    input_context = distribute_lib.InputContext(
        num_input_pipelines=self._num_clients,
        input_pipeline_id=self._client_id,
        num_replicas_in_sync=self._num_replicas_in_sync
    )
    dataset = dataset_fn(input_context)

    # Note that the dataset should already batched to local per-relica batch
    def _create_batch_layout(tensor_spec):
      rank = len(tensor_spec.shape)
      return layout.Layout.batch_sharded(
          self._mesh, batch_dim=dtensor_util.DEFAULT_BATCH_MESH_DIM_NAME,
          rank=rank)

    layouts = nest.map_structure(_create_batch_layout, dataset.element_spec)

    batch_size = distribute.compute_batch_size(dataset)
    # There are multiple case that the batch is not static, eg partial batch,
    # or uneven batch, in all those case, it will return -1.
    if batch_size.numpy() < 0:
      # When we don't have a static batch size.
      raise ValueError('DTensor strategy requires a static batch size for now.'
                       'The dynamic batch size will be supported in future')
    global_batch_size = batch_size.numpy() * self._num_replicas_in_sync

    return input_util.DTensorDataset(
        dataset=dataset,
        mesh=self._mesh,
        layouts=layouts,
        global_batch_size=global_batch_size,
        dataset_already_batched=True,
        batch_dim=dtensor_util.DEFAULT_BATCH_MESH_DIM_NAME,
        # TODO(scottzhu): Add prefetch support by inspecting the input dataset.
        prefetch=None,
        tf_data_service_config=None
    )

  def _experimental_distribute_values_from_function(self, value_fn):
    per_replica_values = []
    # Note that in the multi-worker setting, this function only return the
    # slide of DistributedValue for the current worker.
    for i in range(self._mesh.num_local_devices()):
      # In the case of 2 worker with 2 local devices on each worker,
      # worker 0 will get 0 and 1 for replica_id.
      # worker 1 will get 2 and 3 for replica_id.
      replica_id = d_config.client_id() * self._mesh.num_local_devices() + i
      per_replica_values.append(value_fn(
          distribute_lib.ValueContext(replica_id,
                                      self._num_replicas_in_sync)))
    # Instead of using the DistributeVariable, return a DTensor instead since
    # the run() will expect a DTensor instance.
    result = distribute_utils.regroup(per_replica_values, always_wrap=True)
    map_fn = functools.partial(dtensor_util.convert_per_replica_to_dtensor,
                               mesh=self._mesh)
    return nest.map_structure(map_fn, result)

  def call_for_each_replica(self, fn, args=(), kwargs=None):
    """Run `fn` once per replica.

    This is a method that expected by the strategy base class in its `run()`.

    Args:
      fn: function to run (will be run once per replica).
      args: Tuple or list with positional arguments for `fn`.
      kwargs: Dict with keyword arguments for `fn`.

    Returns:
      Merged return value of `fn` across all replicas.
    """
    # Comparing to the existing MirroredStrategy, which will run the fn on
    # each of the replica with individual thread, the DTensor will just run
    # the fn once with the DTensor inputs, and the distribution will be handled
    # by the DTensor.

    distribute_lib._require_cross_replica_or_default_context_extended(self)   # pylint: disable=protected-access
    if kwargs is None:
      kwargs = {}

    # For any value that is not DTensor, eg normal tf.Tensor or
    # DistributedValues, we need to convert them into DTensor.
    map_fn = functools.partial(dtensor_util.convert_inputs_to_dtensor,
                               mesh=self._mesh)
    d_args = nest.map_structure(map_fn, args)
    d_kwargs = nest.map_structure(map_fn, kwargs)

    with self._container_strategy().scope():
      with dtensor_util.DTensorReplicaContext(self._container_strategy()):
        dtensor_result = fn(*d_args, **d_kwargs)

    return nest.map_structure(
        dtensor_util.DTensorDistributedValue,
        dtensor_result)

  def _gather_to_implementation(self, value, destinations, axis, options):
    if isinstance(value, dtensor_util.DTensorDistributedValue):
      value = value.get_dtensor()
    if not d_api.is_dtensor(value):
      # This is the current behavior for mirrored strategy, should we raise an
      # error for unsupported types?
      return value

    # Unpack the dtensor components and gather the tensors on the axis
    components = d_api.unpack(value)
    return array_ops.concat(components, axis=axis)

  def _use_merge_call(self):
    # This is method for V1 StrategyExtended by still used by
    # tf.__internal__.distribute.strategy_supports_no_merge_call
    return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tensorflow/compiler/mlir/quantization/tensorflow/exported_model.proto
# Protobuf Python Version: 5.28.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    3,
    '',
    'tensorflow/compiler/mlir/quantization/tensorflow/exported_model.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorflow.core.framework import graph_pb2 as tensorflow_dot_core_dot_framework_dot_graph__pb2
from tensorflow.core.protobuf import meta_graph_pb2 as tensorflow_dot_core_dot_protobuf_dot_meta__graph__pb2
from tensorflow.core.protobuf import saver_pb2 as tensorflow_dot_core_dot_protobuf_dot_saver__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nEtensorflow/compiler/mlir/quantization/tensorflow/exported_model.proto\x12\x17tensorflow.quantization\x1a%tensorflow/core/framework/graph.proto\x1a)tensorflow/core/protobuf/meta_graph.proto\x1a$tensorflow/core/protobuf/saver.proto\"\xbe\x03\n\rExportedModel\x12\'\n\tgraph_def\x18\x01 \x01(\x0b\x32\x14.tensorflow.GraphDef\x12\x16\n\x0einit_node_name\x18\x02 \x01(\t\x12\x16\n\x0e\x63heckpoint_dir\x18\x05 \x01(\t\x12U\n\x10\x66unction_aliases\x18\x06 \x03(\x0b\x32;.tensorflow.quantization.ExportedModel.FunctionAliasesEntry\x12\x31\n\x0f\x61sset_file_defs\x18\x08 \x03(\x0b\x32\x18.tensorflow.AssetFileDef\x12\'\n\tsaver_def\x18\n \x01(\x0b\x32\x14.tensorflow.SaverDef\x1a\x36\n\x14\x46unctionAliasesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x07\x10\x08J\x04\x08\t\x10\nR\x15variable_shared_namesR\x11restore_node_nameR\x0esave_node_nameR\x17\x66ile_prefix_tensor_nameb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tensorflow.compiler.mlir.quantization.tensorflow.exported_model_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_EXPORTEDMODEL_FUNCTIONALIASESENTRY']._loaded_options = None
  _globals['_EXPORTEDMODEL_FUNCTIONALIASESENTRY']._serialized_options = b'8\001'
  _globals['_EXPORTEDMODEL']._serialized_start=219
  _globals['_EXPORTEDMODEL']._serialized_end=665
  _globals['_EXPORTEDMODEL_FUNCTIONALIASESENTRY']._serialized_start=504
  _globals['_EXPORTEDMODEL_FUNCTIONALIASESENTRY']._serialized_end=558
# @@protoc_insertion_point(module_scope)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.lite.constants namespace
"""

import sys as _sys

from tensorflow.lite.python.lite_constants import FLOAT # line: 50
from tensorflow.lite.python.lite_constants import FLOAT16 # line: 51
from tensorflow.lite.python.lite_constants import GRAPHVIZ_DOT # line: 60
from tensorflow.lite.python.lite_constants import INT16 # line: 58
from tensorflow.lite.python.lite_constants import INT32 # line: 52
from tensorflow.lite.python.lite_constants import INT64 # line: 53
from tensorflow.lite.python.lite_constants import INT8 # line: 57
from tensorflow.lite.python.lite_constants import QUANTIZED_UINT8 # line: 55
from tensorflow.lite.python.lite_constants import STRING # line: 54
from tensorflow.lite.python.lite_constants import TFLITE # line: 59

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "lite.constants", public_apis=None, deprecation=False,
      has_lite=False)



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.__internal__ namespace
"""

import sys as _sys

from tensorflow._api.v2.compat.v2.__internal__ import autograph
from tensorflow._api.v2.compat.v2.__internal__ import decorator
from tensorflow._api.v2.compat.v2.__internal__ import dispatch
from tensorflow._api.v2.compat.v2.__internal__ import distribute
from tensorflow._api.v2.compat.v2.__internal__ import eager_context
from tensorflow._api.v2.compat.v2.__internal__ import feature_column
from tensorflow._api.v2.compat.v2.__internal__ import function
from tensorflow._api.v2.compat.v2.__internal__ import graph_util
from tensorflow._api.v2.compat.v2.__internal__ import mixed_precision
from tensorflow._api.v2.compat.v2.__internal__ import monitoring
from tensorflow._api.v2.compat.v2.__internal__ import nest
from tensorflow._api.v2.compat.v2.__internal__ import ops
from tensorflow._api.v2.compat.v2.__internal__ import saved_model
from tensorflow._api.v2.compat.v2.__internal__ import smart_cond
from tensorflow._api.v2.compat.v2.__internal__ import test
from tensorflow._api.v2.compat.v2.__internal__ import tf2
from tensorflow._api.v2.compat.v2.__internal__ import tracking
from tensorflow._api.v2.compat.v2.__internal__ import train
from tensorflow._api.v2.compat.v2.__internal__ import types
from tensorflow.python.distribute.load_context import load_context # line: 49
from tensorflow.python.eager.backprop import record_gradient # line: 160
from tensorflow.python.eager.context import is_tfrt_enabled # line: 215
from tensorflow.python.eager.lift_to_graph import lift_to_graph # line: 203
from tensorflow.python.framework.composite_tensor import CompositeTensor # line: 24
from tensorflow.python.framework.func_graph import FuncGraph # line: 133
from tensorflow.python.framework.ops import EagerTensor # line: 674
from tensorflow.python.framework.ops import SymbolicTensor # line: 259
from tensorflow.python.framework.ops import _create_c_op as create_c_op # line: 1020
from tensorflow.python.framework.ops import get_name_scope # line: 5176
from tensorflow.python.ops.control_flow_ops import get_enclosing_xla_context # line: 2208
from tensorflow.python.ops.control_flow_switch_case import execute_fn_for_device # line: 132
from tensorflow.python.util.keras_deps import register_call_context_function # line: 38
from tensorflow.python.util.keras_deps import register_clear_session_function # line: 44
from tensorflow.python.util.keras_deps import register_get_session_function # line: 50
from tensorflow.python.util.keras_deps import register_load_model_function # line: 56



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator2/generator/generator.py script.
"""Public API for tf._api.v2.linalg namespace
"""

import sys as _sys

from tensorflow._api.v2.compat.v1.linalg import experimental
from tensorflow.python.ops.gen_array_ops import matrix_band_part as band_part # line: 4921
from tensorflow.python.ops.gen_array_ops import diag as tensor_diag # line: 1949
from tensorflow.python.ops.gen_linalg_ops import cholesky # line: 766
from tensorflow.python.ops.gen_linalg_ops import matrix_determinant as det # line: 1370
from tensorflow.python.ops.gen_linalg_ops import matrix_inverse as inv # line: 1516
from tensorflow.python.ops.gen_linalg_ops import lu # line: 1252
from tensorflow.python.ops.gen_linalg_ops import qr # line: 2150
from tensorflow.python.ops.gen_linalg_ops import matrix_solve as solve # line: 1694
from tensorflow.python.ops.gen_linalg_ops import matrix_square_root as sqrtm # line: 1913
from tensorflow.python.ops.gen_math_ops import cross # line: 2708
from tensorflow.python.ops.array_ops import matrix_diag as diag # line: 2075
from tensorflow.python.ops.array_ops import matrix_diag_part as diag_part # line: 2244
from tensorflow.python.ops.array_ops import matrix_transpose # line: 1995
from tensorflow.python.ops.array_ops import matrix_set_diag as set_diag # line: 2432
from tensorflow.python.ops.array_ops import tensor_diag_part # line: 2387
from tensorflow.python.ops.array_ops import matrix_transpose as transpose # line: 1995
from tensorflow.python.ops.clip_ops import global_norm # line: 245
from tensorflow.python.ops.linalg.linalg_impl import adjoint # line: 102
from tensorflow.python.ops.linalg.linalg_impl import eigh_tridiagonal # line: 1232
from tensorflow.python.ops.linalg.linalg_impl import matrix_exponential as expm # line: 232
from tensorflow.python.ops.linalg.linalg_impl import logdet # line: 68
from tensorflow.python.ops.linalg.linalg_impl import logm # line: 54
from tensorflow.python.ops.linalg.linalg_impl import lu_matrix_inverse # line: 1035
from tensorflow.python.ops.linalg.linalg_impl import lu_reconstruct # line: 1100
from tensorflow.python.ops.linalg.linalg_impl import lu_solve # line: 937
from tensorflow.python.ops.linalg.linalg_impl import matrix_rank # line: 768
from tensorflow.python.ops.linalg.linalg_impl import pinv # line: 807
from tensorflow.python.ops.linalg.linalg_impl import slogdet # line: 44
from tensorflow.python.ops.linalg.linalg_impl import tridiagonal_matmul # line: 670
from tensorflow.python.ops.linalg.linalg_impl import tridiagonal_solve # line: 446
from tensorflow.python.ops.linalg.linear_operator import LinearOperator # line: 91
from tensorflow.python.ops.linalg.linear_operator_adjoint import LinearOperatorAdjoint # line: 28
from tensorflow.python.ops.linalg.linear_operator_block_diag import LinearOperatorBlockDiag # line: 34
from tensorflow.python.ops.linalg.linear_operator_block_lower_triangular import LinearOperatorBlockLowerTriangular # line: 39
from tensorflow.python.ops.linalg.linear_operator_circulant import LinearOperatorCirculant # line: 740
from tensorflow.python.ops.linalg.linear_operator_circulant import LinearOperatorCirculant2D # line: 1061
from tensorflow.python.ops.linalg.linear_operator_circulant import LinearOperatorCirculant3D # line: 1316
from tensorflow.python.ops.linalg.linear_operator_composition import LinearOperatorComposition # line: 35
from tensorflow.python.ops.linalg.linear_operator_diag import LinearOperatorDiag # line: 32
from tensorflow.python.ops.linalg.linear_operator_full_matrix import LinearOperatorFullMatrix # line: 29
from tensorflow.python.ops.linalg.linear_operator_householder import LinearOperatorHouseholder # line: 32
from tensorflow.python.ops.linalg.linear_operator_identity import LinearOperatorIdentity # line: 101
from tensorflow.python.ops.linalg.linear_operator_identity import LinearOperatorScaledIdentity # line: 538
from tensorflow.python.ops.linalg.linear_operator_inversion import LinearOperatorInversion # line: 25
from tensorflow.python.ops.linalg.linear_operator_kronecker import LinearOperatorKronecker # line: 60
from tensorflow.python.ops.linalg.linear_operator_low_rank_update import LinearOperatorLowRankUpdate # line: 35
from tensorflow.python.ops.linalg.linear_operator_lower_triangular import LinearOperatorLowerTriangular # line: 31
from tensorflow.python.ops.linalg.linear_operator_permutation import LinearOperatorPermutation # line: 35
from tensorflow.python.ops.linalg.linear_operator_toeplitz import LinearOperatorToeplitz # line: 33
from tensorflow.python.ops.linalg.linear_operator_tridiag import LinearOperatorTridiag # line: 39
from tensorflow.python.ops.linalg.linear_operator_zeros import LinearOperatorZeros # line: 39
from tensorflow.python.ops.linalg_ops import cholesky_solve # line: 147
from tensorflow.python.ops.linalg_ops import self_adjoint_eig as eigh # line: 441
from tensorflow.python.ops.linalg_ops import self_adjoint_eigvals as eigvalsh # line: 465
from tensorflow.python.ops.linalg_ops import eye # line: 196
from tensorflow.python.ops.linalg_ops import matrix_solve_ls as lstsq # line: 244
from tensorflow.python.ops.linalg_ops import norm # line: 633
from tensorflow.python.ops.linalg_ops import svd # line: 489
from tensorflow.python.ops.linalg_ops import matrix_triangular_solve as triangular_solve # line: 84
from tensorflow.python.ops.math_ops import matmul # line: 3443
from tensorflow.python.ops.math_ops import matvec # line: 3711
from tensorflow.python.ops.math_ops import tensordot # line: 5215
from tensorflow.python.ops.math_ops import trace # line: 3399
from tensorflow.python.ops.nn_impl import l2_normalize # line: 540
from tensorflow.python.ops.nn_impl import normalize # line: 487
from tensorflow.python.ops.special_math_ops import einsum # line: 618

from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "linalg", public_apis=None, deprecation=False,
      has_lite=False)
